# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import django
from django import forms
from django.contrib import admin
from django.template.defaultfilters import truncatechars
from django.utils.html import mark_safe
from django.utils.translation import ugettext_lazy as _

from ..models import MsgLogFlag, WeChatApp
from ..models.permission import get_user_permissions
from .base import has_wechat_permission


class WeChatAppForm(forms.ModelForm):
    log_message = forms.BooleanField(
        label=_("log messages"), initial=False, required=False)

    wechat_host = forms.CharField(
        label=_("WeChat host"), required=False,
        help_text=_("接收微信回调的域名"))
    wechat_https = forms.BooleanField(
        label=_("WeChat https"), required=False,
        help_text=_("回调地址是否为https"))

    accesstoken_url = forms.URLField(
        label=_("accesstoken url"), required=False,
        help_text=_("获取accesstoken的url,不填直接从微信取"))
    oauth_url = forms.URLField(
        label=_("oauth url"), required=False,
        help_text=_("授权重定向的url,用于第三方网页授权换取code,默认直接微信授权"))

    class Meta(object):
        model = WeChatApp
        fields = "__all__"
        widgets = dict(
            appsecret=forms.PasswordInput(render_value=True),
            token=forms.PasswordInput(render_value=True),
            encoding_aes_key=forms.PasswordInput(render_value=True)
        )

    def __init__(self, *args, **kwargs):
        inst = kwargs.get("instance")
        if inst:
            initial = kwargs.get("initial", {})
            initial["log_message"] = inst.log_message
            initial["wechat_host"] = inst.site_host
            initial["wechat_https"] = inst.site_https
            initial["accesstoken_url"] = inst.configurations.get(
                "ACCESSTOKEN_URL", "")
            initial["oauth_url"] = inst.configurations.get("OAUTH_URL", "")
            kwargs["initial"] = initial
        return super(WeChatAppForm, self).__init__(*args, **kwargs)

    def clean(self):
        cleaned_data = super(WeChatAppForm, self).clean()
        if cleaned_data.get("log_message"):
            cleaned_data["flags"] = MsgLogFlag.LOG_MESSAGE
        else:
            cleaned_data["flags"] = 0
        return cleaned_data

    def save(self, commit=True):
        self.instance.flags = self.cleaned_data["flags"]
        self.instance.configurations["SITE_HOST"] =\
            self.cleaned_data.get("wechat_host", "")
        self.instance.configurations["SITE_HTTPS"] =\
            self.cleaned_data.get("wechat_https", None)
        self.instance.configurations["ACCESSTOKEN_URL"] =\
            self.cleaned_data.get("accesstoken_url", "")
        self.instance.configurations["OAUTH_URL"] =\
            self.cleaned_data.get("oauth_url", "")
        return super(WeChatAppForm, self).save(commit)


@admin.register(WeChatApp)
class WeChatAppAdmin(admin.ModelAdmin):
    change_form_template = "admin/change_form.html"
    change_list_template = "admin/change_list.html"
    delete_confirmation_template = "admin/delete_confirmation.html"
    object_history_template = "admin/object_history.html"

    actions = None
    list_display = (
        "title", "name", "type", "appid", "short_desc", "abilities",
        "created_at", "updated_at")
    search_fields = ("title", "name", "appid", "short_desc")

    fields = (
        "title", "name", "appid", "appsecret", "type", "abilities", "token",
        "encoding_aes_key", "encoding_mode", "desc", "log_message",
        "callback", "wechat_host", "wechat_https", "accesstoken_url",
        "oauth_url", "created_at", "updated_at"
    )
    readonly_fields = ("abilities",)

    @mark_safe
    def abilities(self, obj):
        abilities = []
        if obj.abilities.authed:
            abilities.append(_("authed"))
        if obj.abilities.api:
            abilities.append(_("api"))
        if obj.abilities.interactable:
            abilities.append(_("interactable"))
        if obj.abilities.pay:
            abilities.append(_("WeChat pay"))

        styles = {
            "background-color": "#70be2b",
            "color": "white",
            "border-radius": "5px",
            "margin": "0 2px",
            "padding": "2px 4px"
        }
        style_str = ";".join(["{0}: {1}".format(*o) for o in styles.items()])
        tpl = '<span style="{0}">{1}</span>'
        return "".join(map(lambda o: tpl.format(style_str, o), abilities))
    abilities.short_description = _("abilities")

    def short_desc(self, obj):
        return truncatechars(obj.desc, 35)
    short_desc.short_description = _("description")

    def callback(self, obj):
        return obj and obj.build_url(
            "handler", request=self.request, absolute=True)
    callback.short_description = _("message callback url")

    def delete_view(self, request, object_id, *args, **kwargs):
        request.app = WeChatApp.objects.get(id=object_id)
        return super(WeChatAppAdmin, self).delete_view(
            request, object_id, *args, **kwargs)

    def get_urls(self):
        urlpatterns = super(WeChatAppAdmin, self).get_urls()
        # django 1.11 替换urlpattern为命名式的
        if django.VERSION[0] < 2:
            for pattern in urlpatterns:
                pattern._regex = pattern._regex.replace(
                    "(.+)", "(?P<object_id>.+)")
        return urlpatterns

    def get_fields(self, request, obj=None):
        fields = list(super(WeChatAppAdmin, self).get_fields(request, obj))
        if not obj:
            fields.remove("callback")
            fields.remove("created_at")
            fields.remove("updated_at")
        if obj and obj.type == WeChatApp.Type.SUBSCRIBEAPP:
            fields.remove("oauth_url")
        if obj and not obj.abilities.interactable:
            fields.remove("callback")
        return fields

    def get_readonly_fields(self, request, obj=None):
        rv = super(WeChatAppAdmin, self).get_readonly_fields(request, obj)
        if obj:
            rv = rv + (
                "name", "appid", "callback", "created_at", "updated_at")
            if obj.type != 0:
                rv = rv + ("type",)
        return rv

    def get_queryset(self, request):
        self.request = request
        rv = super(WeChatAppAdmin, self).get_queryset(request)

        # 非管理员 过滤拥有的微信号权限
        if not super(WeChatAppAdmin, self).has_change_permission(request):
            allowed_apps = self._get_allowed_apps(request)
            rv = rv.filter(name__in=allowed_apps)

        return rv

    def has_change_permission(self, request, obj=None):
        rv = super(WeChatAppAdmin, self).has_change_permission(request, obj)
        if not rv:
            if obj:
                return has_wechat_permission(request, obj, "manage")
            else:
                return bool(self._get_allowed_apps(request))
        return rv

    def _get_allowed_apps(self, request):
        """有权限的微信号"""
        perms = get_user_permissions(request.user)
        return set(
            appname for appname, permissions in perms.items()
            if "manage" in permissions
        )

    def get_model_perms(self, request):
        return ({} if getattr(request, "app_id", None)
            else super(WeChatAppAdmin, self).get_model_perms(request))

    def get_deleted_objects(self, objs, request):
        from ..models import WeChatUser
        from ..pay.models import UnifiedOrder
        deleted_objects, model_count, perms_needed, protected =\
            super(WeChatAppAdmin, self).get_deleted_objects(objs, request)
        ignored_models = (
            WeChatUser._meta.verbose_name, UnifiedOrder._meta.verbose_name)
        perms_needed = perms_needed.difference(ignored_models)
        return deleted_objects, model_count, perms_needed, protected

    form = WeChatAppForm



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from ..models import appmethod, WeChatModel
from .base import WeChatTestCase


class ModelBaseTestCase(WeChatTestCase):
    def test_appmethod(self):
        """测试appmethod"""
        class DebugModel(WeChatModel):
            class Meta(object):
                abstract = True

            @classmethod
            @appmethod
            def test(cls, app, *args, **kwargs):
                return cls, app, args, kwargs

            @classmethod
            @appmethod("another")
            def test1(cls, app, *args, **kwargs):
                return cls, app, args, kwargs

        def assertAppMethodEqual(func, app, funcname):
            args = (1, 2)
            kwargs = dict(a=1, b=2)
            cls, app_, args_, kwargs_ = func(*args, **kwargs)
            self.assertEqual(cls, DebugModel)
            self.assertEqual(app_, app)
            self.assertEqual(args_, args)
            self.assertEqual(kwargs_, kwargs)
            self.assertEqual(func.__name__, funcname)
            self.assertEqual(
                func.__qualname__, "{0}.{1}".format("WeChatApp", funcname))

        assertAppMethodEqual(self.app.test, self.app, "test")
        assertAppMethodEqual(self.app.another, self.app, "another")
        assertAppMethodEqual(self.another_app.test, self.another_app, "test")
        assertAppMethodEqual(
            self.another_app.another, self.another_app, "another")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import mimetypes
import re

from django.db import models as m, transaction
from django.utils.translation import ugettext_lazy as _
from wechatpy.constants import WeChatErrorCode
from wechatpy.exceptions import WeChatClientException

from ..utils.model import enum2choices, model_fields
from . import appmethod, WeChatApp, WeChatModel


class MaterialManager(m.Manager):
    def create_material(self, **kwargs):
        """创建永久素材"""
        app = kwargs.pop("app", self.instance)
        type = kwargs.pop("type", None)
        if type is None:
            raise NotImplementedError()
        if type == Material.Type.NEWS:
            return self.create_news(app=app, **kwargs)
        else:
            media_id = kwargs["media_id"]
            if type == Material.Type.VIDEO and "url" not in kwargs:
                data = app.client.material.get(media_id)
                kwargs["url"] = data.get("down_url")

            kwargs = {
                key: kwargs[key]
                for key in model_fields(self.model)
                if key in kwargs
            }
            query = dict(app=app, type=type, media_id=media_id)
            record = dict(app=app, type=type, **kwargs)
            return self.update_or_create(defaults=record, **query)[0]

    def create_news(self, **kwargs):
        """创建永久图文素材"""
        app = kwargs.pop("app", self.instance)
        from . import Article
        # 插入media
        query = dict(app=app, media_id=kwargs["media_id"])
        record = dict(
            type=Material.Type.NEWS,
            update_time=kwargs["update_time"]
        )
        record.update(query)
        news, created = self.update_or_create(defaults=record, **query)
        if not created:
            # 移除所有article重新插入
            news.articles.all().delete()

        articles = (kwargs.get("content") or kwargs)["news_item"]
        fields = model_fields(Article)
        news.articles.bulk_create([
            Article(
                index=idx,
                material=news,
                _thumb_url=article.get("thumb_url"),
                **{k: v for k, v in article.items() if k in fields}
            )
            for idx, article in enumerate(articles)
        ])
        return news


class Material(WeChatModel):
    class Type(object):
        IMAGE = "image"
        VIDEO = "video"
        NEWS = "news"
        VOICE = "voice"

    app = m.ForeignKey(
        WeChatApp, related_name="materials", on_delete=m.CASCADE)

    type = m.CharField(
        _("type"), max_length=5, choices=(enum2choices(Type)))
    media_id = m.CharField(_("media_id"), max_length=64)
    alias = m.CharField(
        _("alias"), max_length=16, blank=True, null=True,
        help_text=_("素材别名,用于程序快速查询素材,单app下唯一"))

    name = m.CharField(_("name"), max_length=64, blank=True, null=True)
    url = m.CharField(_("url"), max_length=512, editable=False, null=True)
    update_time = m.IntegerField(
        _("update time"), editable=False, null=True)

    comment = m.TextField(_("comment"), blank=True)

    created_at = m.DateTimeField(_("created at"), auto_now_add=True)
    updated_at = m.DateTimeField(_("updated at"), auto_now=True)

    objects = MaterialManager()

    class Meta(object):
        verbose_name = _("material")
        verbose_name_plural = _("materials")

        unique_together = (("app", "media_id"), ("app", "alias"))
        ordering = ("app", "-update_time")

    @classmethod
    @appmethod("sync_materials")
    def sync(cls, app, id=None, type=None):
        """同步所有永久素材"""
        if id:
            if type not in (cls.Type.NEWS, cls.Type.VIDEO):
                raise NotImplementedError()
            data = app.client.material.get_raw(id)
            return app.materials.create_material(
                type=type, media_id=id, **data)
        else:
            updated = []
            for type, _ in enum2choices(cls.Type):
                with transaction.atomic():
                    updates = cls.sync_type(app, type)
                    updated.extend(updates)
            return updated

    @classmethod
    @appmethod("sync_type_materials")
    def sync_type(cls, app, type):
        """同步某种类型的永久素材"""
        count = 20
        offset = 0
        updates = []
        while True:
            data = app.client.material.batchget(
                media_type=type,
                offset=offset,
                count=count
            )
            updates.extend(data["item"])
            if data["total_count"] <= offset + count:
                break
            offset += count
        # 删除被删除的 更新或新增获取的
        (app.materials.filter(type=type)
            .exclude(media_id__in=map(lambda o: o["media_id"], updates))
            .delete())
        return [
            app.materials.create_material(type=type, **item)
            for item in updates]

    @classmethod
    def as_permenant(cls, media_id, app, save=True):
        """将临时素材转换为永久素材"""
        # 下载临时素材
        resp = app.client.media.download(media_id)

        try:
            content_type = resp.headers["Content-Type"]
        except:
            raise ValueError("missing Content-Type")
        if content_type.startswith("image"):
            type = cls.Type.IMAGE
        elif content_type.startswith("video"):
            type = cls.Type.VIDEO
        elif content_type.startswith("audio"):
            type = cls.Type.VOICE
        else:
            raise ValueError("unknown Content-Type")

        # 找文件名
        try:
            disposition = resp.headers["Content-Disposition"]
            filename = re.findall(r'filename="(.+?)"', disposition)[0]
        except:
            # 默认文件名
            ext = mimetypes.guess_extension(content_type)
            filename = (media_id + ext) if ext else media_id

        # 上载素材
        return cls.upload_permenant(app, (filename, resp.content), type, save)

    @classmethod
    def upload_permenant(cls, app, file, type, save=True):
        """上传永久素材"""
        data = app.client.material.add(type, file)
        media_id = data["media_id"]
        if save:
            return app.materials.create_material(
                type=type, media_id=media_id, url=data.get("url"))
        else:
            return media_id

    @classmethod
    def upload_temporary(cls, app, file, type):
        """上传临时素材
        :param type: image|voice|video|thumb
        """
        return app.client.media.upload(type, file)

    @property
    def articles_json(self):
        return list(map(lambda o: dict(
            title=o.title,
            description=o.digest,
            image=o.thumb_url,
            url=o.url
        ), self.articles.all()))

    def save(self, *args, **kwargs):
        self.alias = self.alias or None
        return super(Material, self).save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        # 先远程素材删除
        try:
            self.app.client.material.delete(self.media_id)
        except WeChatClientException as e:
            if e.errcode != WeChatErrorCode.INVALID_MEDIA_ID:
                raise
        rv = super(Material, self).delete(*args, **kwargs)
        return rv

    def __str__(self):
        media = "{type}:{media_id}".format(type=self.type, media_id=self.media_id)
        return "{0} ({1})".format(self.comment, media) if self.comment else media



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from uuid import uuid4

from django.utils import timezone as tz
from wechatpy import WeChatPay as WeChatPayBaseClient

from wechat_django.models import WeChatUser
from wechat_django.utils.web import get_ip
from ..models import UnifiedOrder, UnifiedOrderResult, WeChatPay
from ..models.base import PayDateTimeField
from ..signals import order_updated
from .base import mock, WeChatPayTestCase


class OrderTestCase(WeChatPayTestCase):
    def test_create(self):
        """测试生成订单"""
        openid = "openid"

        # 最简订单
        minimal = self.minimal_example
        body = minimal["body"]
        total_fee = minimal["total_fee"]
        user_main = self.app.users.create(openid=openid)
        request = self.rf().get("/")
        order = self.app.pay.create_order(user_main, request, **minimal)
        # 默认值
        self.assertEqual(order.trade_type, UnifiedOrder.TradeType.JSAPI)
        self.assertEqual(order.fee_type, UnifiedOrder.FeeType.CNY)
        self.assertAlmostEqual(
            order.time_start, tz.now(), delta=tz.timedelta(seconds=1))
        self.assertAlmostEqual(
            order.time_expire, tz.now() + tz.timedelta(hours=2),
            delta=tz.timedelta(seconds=1))

        self.assertEqual(order.openid, openid)
        self.assertEqual(order.body, body)
        self.assertEqual(order.total_fee, total_fee)

        # 子商户订单
        user_sub = self.app_sub.users.create(openid=openid)
        order = self.app_sub.pay.create_order(
            user_sub, request, **self.minimal_example)
        self.assertIsNone(order.openid)
        self.assertEqual(order.sub_openid, user_sub.openid)
        order = self.app_sub.pay.create_order(
            user_main, request, **self.minimal_example)
        self.assertIsNone(order.sub_openid)
        self.assertEqual(order.openid, user_main.openid)

        # 完整示例
        full = self.full_example
        order = self.app.pay.create_order(user_main, request, **full)
        for k, v in full.items():
            self.assertEqual(getattr(order, k), v)

    def test_call_args(self):
        """测试调用参数"""
        key_map = dict(
            spbill_create_ip="client_ip",
            openid="user_id",
            sub_openid="sub_user_id"
        )
        full = self.full_example
        full["openid"] = "openid"
        # 将时间转换为字符串
        timezone = tz.pytz.timezone("Asia/Shanghai")
        timeformat = "%Y%m%d%H%M%S"
        request = self.rf().get("/")
        order = self.app.pay.create_order(request=request, **full)
        call_args = order.call_args(request, dt2py=True)
        self.assertEqual(call_args, order.call_args(dt2py=True))
        _call_args = order.call_args(request)
        self.assertEqual(_call_args, order._call_args)
        self.assertEqual(
            call_args["notify_url"],
            self.app.build_url(
                "order_notify", kwargs=dict(payname=self.app.pay.name),
                request=request, absolute=True))
        self.assertEqual(call_args["client_ip"], get_ip(request))
        for k, v in full.items():
            fixed_key = key_map[k] if k in key_map else k
            if k in ("time_start", "time_expire"):
                v = PayDateTimeField.dt2str(v)
            self.assertEqual(_call_args[fixed_key], v)

        # 更新参数
        update = dict(
            client_ip="127.0.0.1",
            total_fee=200
        )
        full = self.full_example
        order = self.app.pay.create_order(request=request, **full)
        call_args = order.call_args(request, **update)
        self.assertEqual(call_args, order._call_args)
        self.assertEqual(call_args, order.call_args())
        for k, v in full.items():
            fixed_key = key_map[k] if k in key_map else k
            if k in ("time_start", "time_expire"):
                v = v.astimezone(timezone).strftime(timeformat)
            fixed_val = update[fixed_key] if fixed_key in update else v
            self.assertEqual(call_args[fixed_key], fixed_val)

    def test_update(self):
        """测试更新订单"""
        # 测试无result更新待支付订单
        with mock.patch.object(UnifiedOrder, "verify"):
            order = self.app.pay.create_order(**self.minimal_example)
            result = self.notpay(self.app.pay, order)
            order.update(result)
            for key, value in result.items():
                if key in UnifiedOrder.ALLOW_UPDATES:
                    v = getattr(order, key)
                    v = v if v is None else str(v)
                    self.assertEqual(v, value)
                if key in self.list_fields(UnifiedOrderResult):
                    v = getattr(order.result, key)
                    v = v if v is None else str(v)
                    self.assertEqual(v, value)
            self.assertEqual(UnifiedOrder.verify.call_count, 1)

        # 测试无result更新已完成订单
        with mock.patch.object(UnifiedOrder, "verify"):
            order = self.app.pay.create_order(**self.minimal_example)
            result = self.success(self.app.pay, order)
            order.update(result)
            for key, value in result.items():
                if key in UnifiedOrder.ALLOW_UPDATES:
                    v = getattr(order, key)
                    v = v if v is None else str(v)
                    self.assertEqual(v, value)
                if key in self.list_fields(UnifiedOrderResult):
                    v = getattr(order.result, key)
                    v = v if v is None else str(v)
                    self.assertEqual(v, value)
            self.assertEqual(UnifiedOrder.verify.call_count, 1)

        # 测试有result更新待支付订单
        order = self.app.pay.create_order(**self.minimal_example)
        result = self.notpay(self.app.pay, order)
        order.update(result)
        with mock.patch.object(UnifiedOrder, "verify"):
            result = self.notpay(self.app.pay, order)
            order.update(result)
            for key, value in result.items():
                if key in UnifiedOrder.ALLOW_UPDATES:
                    v = getattr(order, key)
                    v = v if v is None else str(v)
                    self.assertEqual(v, value)
                if key in self.list_fields(UnifiedOrderResult):
                    v = getattr(order.result, key)
                    v = v if v is None else str(v)
                    self.assertEqual(v, value)
            self.assertEqual(UnifiedOrder.verify.call_count, 1)

        # 测试有result更新已完成订单
        order = self.app.pay.create_order(**self.minimal_example)
        result = self.notpay(self.app.pay, order)
        order.update(result)
        with mock.patch.object(UnifiedOrder, "verify"):
            result = self.success(self.app.pay, order)
            order.update(result)
            for key, value in result.items():
                if key in UnifiedOrder.ALLOW_UPDATES:
                    v = getattr(order, key)
                    v = v if v is None else str(v)
                    self.assertEqual(v, value)
                if key in self.list_fields(UnifiedOrderResult):
                    v = getattr(order.result, key)
                    v = v if v is None else str(v)
                    self.assertEqual(v, value)
            self.assertEqual(UnifiedOrder.verify.call_count, 1)

        # 测试再度更新已完成订单
        with mock.patch.object(UnifiedOrder, "verify"):
            result = self.success(self.app.pay, order)
            order.update(result)
            for key, value in result.items():
                if key in UnifiedOrder.ALLOW_UPDATES:
                    v = getattr(order, key)
                    v = v if v is None else str(v)
                    self.assertEqual(v, value)
                if key in self.list_fields(UnifiedOrderResult):
                    v = getattr(order.result, key)
                    v = v if v is None else str(v)
                    self.assertEqual(v, value)
            self.assertEqual(UnifiedOrder.verify.call_count, 1)

        # 测试回调更新
        result = self.notify(self.app.pay, order)
        order.update(result)
        for key, value in result.items():
            if key in UnifiedOrder.ALLOW_UPDATES:
                v = getattr(order, key)
                v = v if v is None else str(v)
                self.assertEqual(v, value)
            if key in self.list_fields(UnifiedOrderResult):
                v = getattr(order.result, key)
                v = v if v is None else str(v)
                self.assertEqual(v, value)

        # 测试回调更新
        order = self.app.pay.create_order(**self.minimal_example)
        result = self.notify(self.app.pay, order)
        order.update(result)
        for key, value in result.items():
            if key in UnifiedOrder.ALLOW_UPDATES:
                v = getattr(order, key)
                v = v if v is None else str(v)
                self.assertEqual(v, value)
            if key in self.list_fields(UnifiedOrderResult):
                v = getattr(order.result, key)
                v = v if v is None else str(v)
                self.assertEqual(v, value)

    def test_verify(self):
        """测试订单参数是否一致"""
        order = self.app.pay.create_order(**self.minimal_example)
        result = self.success(self.app.pay, order)
        result["total_fee"] = 1
        self.assertRaises(AssertionError, order.verify, result)

        result = self.success(self.app.pay, order)
        result["out_trade_no"] = "123"
        self.assertRaises(AssertionError, order.verify, result)

        result = self.success(self.app.pay, order)
        result["mch_id"] = "abc"
        self.assertRaises(AssertionError, order.verify, result)

        result = self.success(self.app.pay, order)
        order.verify(result)

    def test_signal(self):
        """测试订单状态更新信号"""
        with mock.patch.object(order_updated, "send"):
            order = self.app.pay.create_order(**self.minimal_example)
            result = self.success(self.app.pay, order)
            order.update(result)
            self.assertEqual(order_updated.send.call_count, 1)
            kwargs = dict(
                result=order.result,
                order=order,
                state=UnifiedOrderResult.State.SUCCESS,
                attach=result.get("attach")
            )
            self.assertCallArgsEqual(order_updated.send, kwargs=kwargs)

        with mock.patch.object(order_updated, "send"):
            order = self.app.pay.create_order(**self.minimal_example)
            result = self.success(self.app.pay, order)
            order.update(result, signal=False)
            self.assertEqual(order_updated.send.call_count, 0)

    def test_sync(self):
        """测试同步订单"""
        pass

    def test_prepay(self):
        """测试prepay"""
        pass

    def test_close(self):
        """测试关闭订单"""
        pass

    @property
    def minimal_example(self):
        return dict(
            body="body",
            out_trade_no=str(uuid4()),
            total_fee=101
        )

    @property
    def full_example(self):
        return dict(
            device_info="013467007045764",
            body="body",
            # receipt="Y",
            detail="detail",
            out_trade_no=uuid4(),
            total_fee=101,
            time_start=tz.now(),
            time_expire=tz.now() + tz.timedelta(minutes=30),
            goods_tag="goods_tag",
            trade_type=UnifiedOrder.TradeType.APP,
            product_id="product_id",
            limit_pay=UnifiedOrder.LimitPay.NOCREDIT,
            scene_info={
                "store_info": {
                    "id": "SZTX001",
                    "name": "腾大餐厅",
                    "area_code": "440305",
                    "address": "科技园中一路腾讯大厦"
                }
            }
        )

    def notpay(self, pay, order):
        return {
            "return_code": "SUCCESS",
            "return_msg": "OK",
            "appid": str(pay.appid),
            "mch_id": str(pay.mch_id),
            "device_info": None,
            "nonce_str": str(uuid4()),
            "sign": str(uuid4()),
            "result_code": "SUCCESS",
            "total_fee": "101",
            "out_trade_no": order.out_trade_no,
            "trade_state": "NOTPAY",
            "trade_state_desc": "订单未支付"
        }

    def success(self, pay, order):
        """
        :type pay: wechat_django.pay.models.WeChatPay
        :type order: wechat_django.pay.models.UnifiedOrder
        """
        return {
            "openid": "openid",
            "sub_mch_id": None,
            "cash_fee_type": "CNY",
            "settlement_total_fee": "101",
            "nonce_str": str(uuid4()),
            "return_code": "SUCCESS",
            "err_code_des": "SUCCESS",
            "time_end": "20190613190854",
            "mch_id": str(pay.mch_id),
            "trade_type": "JSAPI",
            "trade_state_desc": "ok",
            "trade_state": "SUCCESS",
            "sign": str(uuid4()),
            "cash_fee": "101",
            "is_subscribe": "Y",
            "return_msg": "OK",
            "fee_type": "CNY",
            "bank_type": "CMC",
            "attach": "sandbox_attach",
            "device_info": "sandbox",
            "out_trade_no": order.out_trade_no,
            "transaction_id": order.out_trade_no,
            "total_fee": "101",
            "appid": str(pay.appid),
            "result_code": "SUCCESS",
            "err_code": "SUCCESS"
        }

    def notify(self, pay, order):
        rv = {
            "appid": pay.appid,
            "bank_type": "CFT",
            "cash_fee": "1",
            "fee_type": "CNY",
            "is_subscribe": "N",
            "mch_id": pay.mch_id,
            "nonce_str": "nonce_str",
            "openid": "openid",
            "out_trade_no": order.out_trade_no,
            "result_code": "SUCCESS",
            "return_code": "SUCCESS",
            "time_end": "20190618223614",
            "total_fee": "101",
            "trade_type": "JSAPI",
            "transaction_id": order.out_trade_no,
            "sign": "sign"
        }
        rv["trade_state"] = rv["result_code"]
        return rv



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import time

from django.urls import reverse
from django.utils.http import urlencode
from wechatpy.replies import deserialize_reply, TextReply
from wechatpy.utils import WeChatSigner

from .. import settings
from ..models import MessageHandler, Reply, Rule
from .base import WeChatTestCase

# TODO: 应该拆解成测试各方法会比较直观


class HandlerTestCase(WeChatTestCase):
    def setUp(self):
        super(HandlerTestCase, self).setUp()
        MessageHandler.objects.create_handler(
            app=self.app,
            rules=[Rule(
                type=Rule.Type.EQUAL,
                pattern=self.match_str
            )],
            replies=[Reply(
                type=Reply.MsgType.TEXT,
                content=self.success_reply
            )]
        )

        settings.MESSAGENOREPEATNONCE = False

    def test_badrequest(self):
        """测试错误请求"""
        nonce = "123456"

        # 错误时间戳
        query = dict(
            timestamp=str(int(time.time() - 1000)),
            nonce=nonce
        )
        resp = self.post(query)
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.content, b"")
        settings.MESSAGETIMEOFFSET = 2000
        resp = self.post(query)
        self.assertEqual(resp.status_code, 200)
        reply = deserialize_reply(resp.content)
        self.assertEqual(reply.content, self.success_reply)

        # 错误签名
        query["signature"] = "666666"
        resp = self.post(query)
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.content, b"")

        # 错误格式
        del query["signature"]
        resp = self.post(query)
        self.assertEqual(resp.status_code, 200)
        reply = deserialize_reply(resp.content)
        self.assertEqual(reply.content, self.success_reply)
        resp = self.client.post(self.url + "?" + urlencode(query), dict(a=1))
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.content, b"")

        # 防重放
        settings.MESSAGENOREPEATNONCE = True
        resp = self.post(query)
        self.assertEqual(resp.status_code, 200)
        reply = deserialize_reply(resp.content)
        self.assertEqual(reply.content, self.success_reply)
        resp = self.post(query)
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.content, b"")

        # 防重放超时正常接收
        pass

    def test_request(self):
        """测试正常请求"""
        timestamp = str(int(time.time()))
        nonce = "123456"
        query = dict(
            timestamp=timestamp,
            nonce=nonce
        )

        # 未匹配到
        resp = self.post(query, "666")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.content, b"")

        # 正常匹配
        resp = self.post(query)
        self.assertEqual(resp.status_code, 200)
        reply = deserialize_reply(resp.content)
        self.assertEqual(reply.target, self.sender)
        self.assertIsInstance(reply, TextReply)
        self.assertEqual(reply.content, self.success_reply)

    def test_echostr(self):
        """测试初次请求验证"""
        echostr = b"666666"
        timestamp = str(int(time.time()))
        nonce = "123456"
        query = dict(
            timestamp=timestamp,
            nonce=nonce,
            echostr=echostr
        )
        query["signature"] = self.sign(query)
        resp = self.client.get(self.url, query)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.content, echostr)

    def sign(self, query):
        signer = WeChatSigner()
        signer.add_data(
            self.app.token,
            query["timestamp"],
            query["nonce"]
        )
        return signer.signature

    def post(self, query, content=""):
        xml = """<xml>
        <ToUserName><![CDATA[toUser]]></ToUserName>
        <FromUserName><![CDATA[{sender}]]></FromUserName>
        <CreateTime>{timestamp}</CreateTime>
        <MsgType><![CDATA[text]]></MsgType>
        <Content><![CDATA[{content}]]></Content>
        <MsgId>1234567890123456</MsgId>
        </xml>""".format(
            sender=self.sender,
            content=content or self.match_str,
            timestamp=query["timestamp"]
        )
        if "signature" not in query:
            query["signature"] = self.sign(query)
        return self.client.generic("POST", self.url + "?" + urlencode(query),
            xml)

    @property
    def sender(self):
        return "sender"

    @property
    def match_str(self):
        return "abc"

    @property
    def success_reply(self):
        return "success"

    @property
    def url(self):
        return reverse(
            "wechat_django:handler", kwargs=dict(appname=self.app.name))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models as m
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from django.utils.translation import ugettext_lazy as _

from wechat_django.models import WeChatApp
from wechat_django.utils.func import Static
from .. import settings
from ..client import WeChatPayClient


class WeChatPay(m.Model):
    app = m.ForeignKey(
        WeChatApp, on_delete=m.CASCADE, related_name="pays")

    title = m.CharField(
        _("title"), max_length=16, blank=True,
        help_text=_("商户号标识,用于后台辨识商户号"))
    name = m.CharField(
        _("name"), max_length=16, null=False, default=_("default"),
        help_text=_("商户号程序标识"))
    weight = m.IntegerField(_("weight"), default=0, null=False)

    mch_id = m.CharField(
        _("mch_id"), max_length=32, help_text=_("微信支付分配的商户号"))
    api_key = m.CharField(
        _("WeChatPay api_key"), max_length=128, help_text=_("商户号key"))

    sub_mch_id = m.CharField(
        _("sub_mch_id"), max_length=32, blank=True, null=True,
        help_text=_("子商户号，受理模式下填写"))
    mch_app_id = m.CharField(
        _("mch_app_id"), max_length=32,
        help_text=_("微信分配的主商户号appid，受理模式下填写"),
        blank=True, null=True)

    mch_cert = m.BinaryField(_("mch_cert"), blank=True, null=True)
    mch_key = m.BinaryField(_("mch_key"), blank=True, null=True)

    @property
    def appid(self):
        return self.mch_app_id if self.mch_app_id else self.app.appid

    @property
    def sub_appid(self):
        return self.app.appid if self.mch_app_id else None

    @cached_property
    def staticname(self):
        """发送信号的sender"""
        if not self.pk:
            raise AttributeError
        return Static("{appname}.{payname}".format(
            appname=self.app.name, payname=self.name))

    class Meta(object):
        verbose_name = _("WeChat pay")
        verbose_name_plural = _("WeChat pay")
        unique_together = (("app", "name"),)
        ordering = ("app", "-weight", "pk")

    @property
    def client(self):
        """:rtype: wechat_django.client.WeChatPayClient"""
        if not hasattr(self, "_client"):
            client_factory = import_string(settings.PAYCLIENT)
            self._client = client_factory(self)
        return self._client

    def __str__(self):
        return "{0} ({1})".format(self.title, self.name)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json

from django.conf.urls import url
from django.http import response
from django.test import RequestFactory
from django.test.utils import override_settings
from django.utils.http import urlencode
from six.moves.urllib.parse import parse_qsl, urlparse

from ..models import WeChatUser
from ..oauth import (
    wechat_auth, WeChatSNSScope, WeChatOAuthClient, WeChatOAuthView)
from .base import mock, WeChatTestCase
from .interceptors import common_interceptor


def oauth_api(openid, callback=None):
    kwargs = dict(
        netloc=r"(.*\.)?api\.weixin\.qq\.com$",
        path="/sns/oauth2/access_token"
    )

    def mock(url, request):
        headers = {
            "Content-Type": "application/json"
        }
        data = dict(
            access_token="123",
            refresh_token="123",
            expires_in=7200,
            openid=openid
        )
        from httmock import response
        resp = response(200, data, headers)
        if callback:
            callback(url, request, resp)
        return resp

    return common_interceptor(mock, **kwargs)


def snsinfo_api(openid, callback=None, **data):
    kwargs = dict(
        netloc=r"(.*\.)?api\.weixin\.qq\.com$",
        path="/sns/userinfo"
    )

    def mock(url, request):
        headers = {
            "Content-Type": "application/json"
        }
        from httmock import response
        resp = response(200, data, headers)
        if callback:
            callback(url, request, resp)
        return resp

    return common_interceptor(mock, **kwargs)


@wechat_auth("test", state="state")
def test_oauth_view(request, *args):
    user = request.wechat.user
    return dict(
        openid=user.openid,
        id=user.id,
        args=args
    )

urlpatterns = [
    url(r"^test/(.+)$", test_oauth_view)
]


@override_settings(ROOT_URLCONF=__name__)
class OAuthTestCase(WeChatTestCase):
    def test_oauth_api(self):
        """测试oauth接口请求"""
        pass

    def test_redirect_uri(self):
        """测试redirect_uri生成"""
        host = "example.com"
        rf = RequestFactory(HTTP_HOST=host)
        api = "/app"
        absolute_uri = lambda uri: "http://" + host + uri

        # 测试写入redirect_uri时
        redirect_uri = "https://mp.weixin.qq.com/wiki"
        decorator = wechat_auth(self.app.name, redirect_uri=redirect_uri)
        request = rf.get(api)
        self.assertEqual(decorator.redirect_uri(request), redirect_uri)

        redirect_uri = "/app"
        decorator = wechat_auth(self.app.name, redirect_uri=redirect_uri)
        request = rf.get(api)
        self.assertEqual(
            decorator.redirect_uri(request),
            absolute_uri(redirect_uri)
        )

        # 测试模板请求redirect_uri
        decorator = wechat_auth(self.app.name)
        request = rf.get(redirect_uri)
        self.assertEqual(
            decorator.redirect_uri(request),
            absolute_uri(redirect_uri)
        )

        # 测试ajax请求redirect_uri
        request = rf.post(api, dict(a=1),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
            HTTP_REFERRER=absolute_uri(redirect_uri))
        self.assertEqual(
            decorator.redirect_uri(request),
            absolute_uri(redirect_uri)
        )

    def test_unauthorization(self):
        rf = RequestFactory()
        path = "/test"
        redirect_uri = "https://mp.weixin.qq.com/wiki"

        # 默认response
        state = "abc"
        request = rf.get(path)
        handler = self._make_handler(
            request, redirect_uri=redirect_uri,
            scope=WeChatSNSScope.USERINFO, state=state)
        resp = handler.unauthorization_response(request)
        self.assertIsInstance(resp, response.HttpResponseRedirect)
        client = WeChatOAuthClient(self.app)
        self.assertEqual(resp.url, client.authorize_url(
            redirect_uri, WeChatSNSScope.USERINFO, state))

        # 传入callable state
        state = lambda request: request.path
        request = rf.get(path)
        handler = self._make_handler(
            request, redirect_uri=redirect_uri,
            scope=WeChatSNSScope.USERINFO, state=state)
        resp = handler.unauthorization_response(request)
        self.assertIsInstance(resp, response.HttpResponseRedirect)
        client = WeChatOAuthClient(self.app)
        self.assertEqual(resp.url, client.authorize_url(
            redirect_uri, WeChatSNSScope.USERINFO, path))

        # 传入callback response
        resp = response.HttpResponseForbidden()
        request = rf.get(path)
        handler = self._make_handler(request, response=resp)
        self.assertIs(
            resp,
            handler.unauthorization_response(request)
        )

        # 传入普通response
        resp_func = lambda request: response.HttpResponseForbidden(request.path)
        request = rf.get(path)
        handler = self._make_handler(request, response=resp_func)
        resp = handler.unauthorization_response(request)
        self.assertIsInstance(resp, response.HttpResponseForbidden)
        self.assertEqual(resp.content, path.encode())

    def test_auth(self):
        """测试授权"""
        openid = "abc"
        code = "def"

        # base授权
        request = self.rf().get("/test?code=" + code)
        handler = self._make_handler(request)
        with oauth_api(openid):
            user = handler.auth(self.app, code)
            self.assertTrue(user.id)
            self.assertEqual(user.openid, openid)

        # snsinfo授权
        handler = self._make_handler(request, scope=WeChatSNSScope.USERINFO)
        with oauth_api(openid), snsinfo_api(openid, nickname=code):
            user = handler.auth(self.app, code)
            self.assertTrue(user.id)
            self.assertEqual(user.openid, openid)
            self.assertEqual(user.nickname, code)

    def test_session(self):
        """测试授权后session状态保持"""
        openid = "12345"
        rf = RequestFactory()
        request = rf.get("/test?code=123")
        view = lambda request, *args, **kwargs: response.HttpResponse(request.wechat.openid)
        handler = self._make_handler(request, view=view)
        with oauth_api(openid):
            resp = handler.dispatch(request)
            self.assertEqual(resp.content, openid.encode())

        # 首次授权后不再授权
        session_key = "wechat_test_user"

        def ban_api(*args, **kwargs):
            self.assertFalse(True)
        request = rf.get("/test?code=123")
        request.COOKIES[session_key] = resp.cookies[session_key].value
        with oauth_api(openid, ban_api):
            resp = handler.dispatch(request)
        request = rf.get("/")
        request.COOKIES[session_key] = resp.cookies[session_key].value
        with oauth_api(openid, ban_api):
            resp = handler.dispatch(request)

    def test_user_update(self):
        "测试更新授权数据"
        pass

    def test_request(self):
        "测试请求"
        redirect_uri = "https://mp.weixin.qq.com/wiki"
        host = "example.com"
        url = "/test"
        rf = RequestFactory(HTTP_HOST=host)
        request = rf.get(url)

        # 设置了response
        resp = response.HttpResponseForbidden()
        handler = self._make_handler(
            request, redirect_uri=redirect_uri, response=resp)
        self.assertIs(handler.dispatch(request), resp)

        # 未授权
        handler = self._make_handler(request, redirect_uri=redirect_uri)
        resp = handler.dispatch(request)
        self.assertIsInstance(resp, response.HttpResponseRedirect)
        self.assertEqual(resp.url, request.wechat.oauth_uri)

        # 授权
        request = rf.get(url + "?code=123")
        openid = "456"
        view = lambda request, *args, **kwargs: request.wechat.openid
        handler = self._make_handler(request, redirect_uri=redirect_uri, view=view)
        with oauth_api(openid):
            resp = handler.dispatch(request)
            self.assertEqual(resp.content, openid.encode())

        # 已授权
        session_key = "wechat_test_user"
        request.COOKIES[session_key] = resp.cookies[session_key].value
        resp = handler.dispatch(request)
        self.assertEqual(resp.content, openid.encode())

    def test_classbased_view(self):
        """测试class based view与decorate是否一致"""
        def make_class(attrs=None, initKwargs=None):
            try:
                cls = type("test", (WeChatOAuthView,), attrs or {})
            except:
                cls = type(b"test", (WeChatOAuthView,), attrs or {})
            return cls(**initKwargs or {})

        def assertAuthInfoEqual(**kwargs):
            class_based = make_class(attrs=kwargs)
            class_based_init = make_class(initKwargs=kwargs)
            decorator_based = wechat_auth(**kwargs)(lambda request: None)
            oauth_info = decorator_based.oauth_info
            c_info = class_based.oauth_info
            ci_info = class_based_init.oauth_info
            self.assertEqual(oauth_info.appname, c_info.appname)
            self.assertEqual(oauth_info.scope, c_info.scope)
            self.assertEqual(oauth_info._redirect_uri, c_info._redirect_uri)
            self.assertEqual(oauth_info.required, c_info.required)
            self.assertEqual(oauth_info.response, c_info.response)
            self.assertEqual(oauth_info.state, c_info.state)

            self.assertEqual(oauth_info.appname, ci_info.appname)
            self.assertEqual(oauth_info.scope, ci_info.scope)
            self.assertEqual(oauth_info._redirect_uri, ci_info._redirect_uri)
            self.assertEqual(oauth_info.required, ci_info.required)
            self.assertEqual(oauth_info.response, ci_info.response)
            self.assertEqual(oauth_info.state, ci_info.state)

        assertAuthInfoEqual(appname=self.app.name)
        assertAuthInfoEqual(
            appname=self.app.name, scope=WeChatSNSScope.USERINFO)
        assertAuthInfoEqual(
            appname=self.app.name, redirect_uri="https://mp.weixin.qq.com/")
        assertAuthInfoEqual(appname=self.app.name, state="state")
        assertAuthInfoEqual(appname=self.app.name, required=False)
        assertAuthInfoEqual(
            appname=self.app.name, response=response.HttpResponse())

    def test_view(self):
        """测试view是否正常"""
        args = "t"
        openid = "666"
        host = "example.com"
        path = "/test/" + args

        resp = self.client.get(path, HTTP_HOST="example.com")
        self.assertEqual(resp.status_code, 302)
        parsed_url = urlparse(resp.url)
        self.assertEqual(parsed_url.netloc, "open.weixin.qq.com")
        query = dict(parse_qsl(parsed_url.query))
        self.assertEqual(query["appid"], self.app.appid)
        self.assertEqual(query["response_type"], "code")
        self.assertEqual(query["scope"], WeChatSNSScope.BASE)
        self.assertEqual(query["state"], "state")
        self.assertEqual(parsed_url.fragment, "wechat_redirect")
        self.assertEqual(query["redirect_uri"], "http://" + host + path)

        with oauth_api(openid):
            resp = self.client.get(path + "?code=123", follow=True,
                HTTP_HOST=host)
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.content.decode())
        self.assertEqual(data["openid"], openid)
        self.assertEqual(data["args"], [args])
        id = data["id"]
        user = WeChatUser.objects.get(id=id)
        self.assertEqual(user.openid, openid)

    def test_custom_oauth_url(self):
        """测试设置了OAUTH_URL url后 不再向微信请求授权,转向第三方请求授权"""
        with mock.patch.object(WeChatOAuthClient, "authorize_url"):
            new_url = "new_url"
            self.app.configurations["OAUTH_URL"] = new_url
            hasattr(self.app, "_oauth") and delattr(self.app, "_oauth")
            url = self.app.oauth.authorize_url("redirect_url")
            self.assertTrue(url.startswith(new_url))
            delattr(self.app, "_oauth")
            del self.app.configurations["OAUTH_URL"]

    def _make_handler(self, request, appname="", **kwargs):
        view = kwargs.pop("view", lambda request, *args, **kwargs: "")
        appname = appname or self.app.name
        decorator = wechat_auth(appname, **kwargs)
        handler = WeChatOAuthView(decorator, view)
        handler._patch_request(request)
        return handler



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging

from django.db import models as m
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from django.utils.translation import ugettext_lazy as _
from jsonfield import JSONField
import six
from wechatpy.crypto import WeChatCrypto
from wechatpy.constants import WeChatErrorCode
from wechatpy.exceptions import WeChatClientException

from .. import settings
from ..exceptions import WeChatAbilityError
from ..utils.model import enum2choices
from . import MsgLogFlag
from .ability import Abilities
from ..constants import AppType


class WeChatAppQuerySet(m.QuerySet):
    def get_by_name(self, name):
        return self.get(name=name)


class WeChatAppManager(m.Manager.from_queryset(WeChatAppQuerySet)):
    pass


class WeChatApp(m.Model):
    @staticmethod
    def __new__(cls, *args, **kwargs):
        self = super(WeChatApp, cls).__new__(cls)
        self.abilities = Abilities(self)
        return self

    class EncodingMode(object):
        PLAIN = 0
        # BOTH = 1
        SAFE = 2

    class Flag(object):
        UNAUTH = 0x01 # 未认证

    Type = AppType # TODO: deprecated

    title = m.CharField(
        _("title"), max_length=16, null=False,
        help_text=_("公众号名称,用于后台辨识公众号"))
    name = m.CharField(
        _("name"), max_length=16, blank=False, null=False, unique=True,
        help_text=_("公众号唯一标识,可采用微信号,设定后不可修改,用于程序辨识"))
    desc = m.TextField(_("description"), default="", blank=True)

    appid = m.CharField(_("AppId"), max_length=32, null=False)
    appsecret = m.CharField(_("AppSecret"), max_length=64, blank=True, null=True)
    type = m.PositiveSmallIntegerField(_("type"), default=Type.SERVICEAPP,
        choices=enum2choices(Type))

    token = m.CharField(max_length=32, null=True, blank=True)
    encoding_aes_key = m.CharField(
        _("EncodingAESKey"), max_length=43, null=True, blank=True)
    encoding_mode = m.PositiveSmallIntegerField(
        _("encoding mode"), default=EncodingMode.PLAIN,
        choices=enum2choices(EncodingMode))

    flags = m.IntegerField(_("flags"), default=0)

    ext_info = JSONField(db_column="ext_info", default={})
    configurations = JSONField(db_column="configurations", default={})

    created_at = m.DateTimeField(_("created at"), auto_now_add=True)
    updated_at = m.DateTimeField(_("updated at"), auto_now=True)

    objects = WeChatAppManager()

    abilities = Abilities()

    @cached_property
    def pay(self):
        return self.pays.first()

    class Meta(object):
        verbose_name = _("WeChat app")
        verbose_name_plural = _("WeChat apps")

    @property
    def site_https(self):
        return self.configurations.get("SITE_HTTPS", settings.SITE_HTTPS)

    @property
    def site_host(self):
        return self.configurations.get("SITE_HOST", settings.SITE_HOST)

    @property
    def type_name(self):
        if self.type == WeChatApp.Type.MINIPROGRAM:
            return _("MINIPROGRAM")
        elif self.type == WeChatApp.Type.SERVICEAPP:
            return _("SERVICEAPP")
        elif self.type == WeChatApp.Type.SUBSCRIBEAPP:
            return _("SUBSCRIBEAPP")
        else:
            return _("unknown")

    @property
    def log_message(self):
        return bool(self.flags & MsgLogFlag.LOG_MESSAGE)

    @property
    def log_reply(self):
        return bool(self.flags & MsgLogFlag.LOG_REPLY)

    @property
    def client(self):
        """
        :rtype: wechat_django.client.WeChatClient
                or wechatpy.client.api.wxa.WeChatWxa
        """
        if not self.abilities.api:
            raise WeChatAbilityError(WeChatAbilityError.API)
        if not hasattr(self, "_client"):
            client_factory = import_string(settings.WECHATCLIENT)
            self._client = client_factory(self)
            if self.type == self.Type.MINIPROGRAM:
                self._client = self._client.wxa
        return self._client

    @property
    def oauth(self):
        """
        :rtype: wechat_django.WeChatOAuthClient
        """
        if not self.abilities.oauth:
            raise WeChatAbilityError(WeChatAbilityError.OAUTH)
        if not hasattr(self, "_oauth"):
            client_factory = import_string(settings.OAUTHCLIENT)
            self._oauth = client_factory(self)
        return self._oauth

    @property
    def crypto(self):
        if not self.abilities.interactable:
            raise WeChatAbilityError(WeChatAbilityError.INTERACTABLE)
        if self.encoding_mode != self.EncodingMode.SAFE:
            return
        if not hasattr(self, "_crypto"):
            self._crypto = WeChatCrypto(
                self.token,
                self.encoding_aes_key,
                self.appid
            )
        return self._crypto

    def auth(self, code, scope=None):
        """用code进行微信授权
        :rtype: (wechat_django.models.WeChatUser, dict)
        :raises: wechatpy.exceptions.WeChatClientException
        :raises: wechatpy.exceptions.WeChatOAuthException
        """
        if not self.abilities.api:
            raise WeChatAbilityError(WeChatAbilityError.API)

        if self.type == WeChatApp.Type.SERVICEAPP:
            if not self.abilities.oauth:
                raise WeChatAbilityError(WeChatAbilityError.OAUTH)
            return self._auth_service(code, scope)
        elif self.type == WeChatApp.Type.MINIPROGRAM:
            return self._auth_miniprogram(code)
        else:
            raise WeChatClientException(WeChatErrorCode.UNAUTHORIZED_API)

    def _auth_service(self, code, scope=None):
        """服务号授权
        :rtype: (wechat_django.models.WeChatUser, dict)
        :raises: wechatpy.exceptions.WeChatOAuthException
        """
        from ..oauth import WeChatSNSScope
        if isinstance(scope, six.text_type):
            scope = (scope,)
        data = self.oauth.fetch_access_token(code)
        if scope and WeChatSNSScope.USERINFO in scope:
            # TODO: 优化授权流程 记录accesstoken及refreshtoken 延迟取userinfo
            user_info = self.oauth.get_user_info()
            data.update(user_info)
        return self.users.upsert_by_dict(data), data

    def _auth_miniprogram(self, code):
        """小程式授权
        :rtype: (wechat_django.models.WeChatUser, dict)
        :raises: wechatpy.exceptions.WeChatClientException
        """
        from . import Session
        data = self.client.code_to_session(code)
        user = self.users.upsert_by_dict(data)
        # 持久化session_key
        user.sessions.all().delete()
        user.sessions.add(Session(
            type=Session.Type.MINIPROGRAM,
            auth=dict(session_key=data["session_key"])
        ), bulk=False)
        # 移除session缓存
        try:
            del user.session
        except AttributeError:
            pass
        return user, data

    def build_url(self, urlname, kwargs=None, request=None, absolute=False):
        """构建url"""
        kwargs = kwargs or dict()
        kwargs["appname"] = self.name
        location = reverse("wechat_django:{0}".format(urlname), kwargs=kwargs)
        if not absolute:
            return location

        if request and not self.site_host:
            baseurl = "{}://{}".format(
                "https" if self.site_https else request.scheme,
                request.get_host())
        else:
            protocol = "https://" if self.site_https else "http://"
            if self.site_host:
                host = self.site_host
            else:
                allowed_hosts = settings.settings.ALLOWED_HOSTS
                if not allowed_hosts:
                    raise RuntimeError(
                        "You need setup a WECHAT_SITE_HOST when build "
                        "absolute url.")
                host = allowed_hosts[0]
            baseurl = protocol + host
        return baseurl + location

    def logger(self, name):
        return logging.getLogger(
            "wechat.{name}.{app}".format(name=name, app=self.name))

    def __str__(self):
        rv = "{title} ({name}) - {type}".format(
            title=self.title, name=self.name, type=self.type_name)
        if six.PY2:
            rv = rv.encode("utf-8")
        return rv



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from wechatpy.client.api import WeChatMenu

from ..models import Material, Menu, WeChatApp
from .base import mock, WeChatTestCase


class MenuTestCase(WeChatTestCase):
    def test_sync(self):
        """测试同步菜单"""
        permenant_media = "permenant_media_id"

        app = self.app

        # 微信官方菜单
        data = self.load_data("mp_menu_data")
        buttons = data["selfmenu_info"]["button"]
        with mock.patch.object(Material, "as_permenant"),\
            mock.patch.object(WeChatApp, "sync_articles"),\
            mock.patch.object(WeChatMenu, "get_menu_info"):

            Material.as_permenant.return_value = permenant_media
            WeChatApp.sync_articles.return_value = None
            WeChatMenu.get_menu_info.return_value = data

            app.sync_menus()
            self.assertMenusEqual(self.menus, buttons)

        # 微信菜单不存在
        data = {
            "is_menu_open": 0
        }
        with mock.patch.object(WeChatMenu, "get_menu_info"):
            WeChatMenu.get_menu_info.return_value = data

            Menu.sync(app)
            self.assertEqual(self.menus.count(), 0)

        # 同步自定义菜单
        data = self.load_data("self_menu_data")
        buttons = data["selfmenu_info"]["button"]
        with mock.patch.object(WeChatMenu, "get_menu_info"):
            WeChatMenu.get_menu_info.return_value = data

            Menu.sync(app)
            self.assertMenusEqual(self.menus, buttons)

    def test_menu_publish(self):
        # 菜单发布
        pass

    def assertMenusEqual(self, menus, buttons):
        self.assertEqual(len(menus), len(buttons))
        for menu, button in zip(menus, buttons):
            self.assertMenuEqual(menu, button)

    def assertMenuEqual(self, menu, button):
        self.assertEqual(menu.name, button["name"])
        if "sub_button" in button:
            self.assertIsNone(menu.type)
            sub_menus = menu.sub_button.all()
            sub_buttons = button["sub_button"]["list"]
            self.assertMenusEqual(sub_menus, sub_buttons)
        elif button["type"] == Menu.Event.CLICK:
            self.assertEqual(menu.type, button["type"])
            self.assertEqual(menu.content["key"], button["key"])
        elif button["type"] == Menu.Event.VIEW:
            self.assertEqual(menu.type, button["type"])
            self.assertEqual(menu.content["url"], button["url"])
        else:
            # 回复转换为handler
            pass

    @property
    def menus(self):
        return (self.app.menus.filter(menuid__isnull=True)
            .filter(parent_id__isnull=True)
            .all())



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.test import override_settings
from django.urls import reverse
from wechatpy.client import WeChatClient as _Client
from wechatpy.client.api import WeChatWxa

from ..models import WeChatApp
from .. import settings
from .base import mock, WeChatTestCase
from .interceptors import wechatapi, wechatapi_accesstoken, wechatapi_error


class AppTestCase(WeChatTestCase):
    def test_getaccesstoken(self):
        """测试accesstoken获取"""
        api = "/cgi-bin/token"
        # 测试获取accesstoken
        with wechatapi_accesstoken(lambda url, req, resp: self.assertTrue(req)):
            token = self.app.client.access_token
            self.assertEqual(token, "ACCESS_TOKEN")
        # 测试获取后不再请求accesstoken
        success = dict(
            errcode=0,
            errmsg=""
        )
        with wechatapi_error(api), wechatapi("/cgi-bin/message/custom/send", success):
            resp = self.app.client.message.send_text("openid", "abc")
            self.assertEqual(resp["errcode"], 0)

    @override_settings(CACHES={
        "default": {
            'BACKEND': 'django.core.cache.backends.dummy.DummyCache',
        }
    })
    def test_custom_accesstoken_url(self):
        """测试设置了ACCESSTOKEN url后 不再向原地址发送请求,转为向新地址发送请求"""
        with mock.patch.object(_Client, "_fetch_access_token"):
            new_url = "new_url"
            self.app.configurations["ACCESSTOKEN_URL"] = new_url
            hasattr(self.app, "_client") and delattr(self.app, "_client")
            self.app.client.access_token
            self.assertEqual(
                _Client._fetch_access_token.call_args[0][0], new_url)
            delattr(self.app, "_client")
            del self.app.configurations["ACCESSTOKEN_URL"]

    def test_miniprogram_auth(self):
        """测试小程序授权"""
        openid = "mini_openid"
        session_key = "session_key"
        unionid = "unionid"
        app = self.miniprogram
        with mock.patch.object(WeChatWxa, "code_to_session"):
            return_value = dict(
                openid=openid,
                session_key=session_key,
                unionid=unionid
            )
            WeChatWxa.code_to_session.return_value = return_value
            code = "abcabc"
            user, data = app.auth(code)
            user_id = user.id
            WeChatWxa.code_to_session.assert_called_with(code)
            self.assertEqual(data, return_value)
            self.assertEqual(user.openid, openid)
            self.assertEqual(user.unionid, unionid)
            self.assertEqual(user.sessions.count(), 1)
            self.assertEqual(user.session.session_key, session_key)

        # 再一次授权
        session_key = "another_key"
        with mock.patch.object(WeChatWxa, "code_to_session"):
            return_value = dict(
                openid=openid,
                session_key=session_key,
                unionid=unionid
            )
            WeChatWxa.code_to_session.return_value = return_value
            code = "dddddd"
            user, data = app.auth(code)
            user_id = user.id
            WeChatWxa.code_to_session.assert_called_with(code)
            self.assertEqual(data, return_value)
            self.assertEqual(user.id, user_id)
            self.assertEqual(user.openid, openid)
            self.assertEqual(user.unionid, unionid)
            self.assertEqual(user.sessions.count(), 1)
            self.assertEqual(user.session.session_key, session_key)

    def test_build_url(self):
        """测试url构建"""
        def assertUrlCorrect(hostname, urlname, request=None, secure=False, kwargs=None):
            kwargs_copy = (kwargs or dict()).copy()
            kwargs_copy["appname"] = self.app.name
            location = reverse(
                "wechat_django:{0}".format(urlname), kwargs=kwargs_copy)
            self.assertEqual(
                location, self.app.build_url(urlname, kwargs, absolute=False))

            protocol = "https://" if secure else "http://"
            self.assertEqual(
                protocol + hostname + location,
                self.app.build_url(
                    urlname, kwargs, request=request, absolute=True))

        url_name = "handler"
        allowed_host = "example.com"
        request_host = "example.com:1000"
        settings_host = "example.com:2000"
        configure_host = "example.com:3000"
        req = self.rf(SERVER_NAME=request_host).get("/")

        # 连request都没有时,取allowed host的第一个
        assertUrlCorrect(allowed_host, url_name, secure=True)

        # 什么都没有配置时,取request的host
        assertUrlCorrect(request_host, url_name, req, secure=True)

        settings.SITE_HOST = settings_host
        settings.SITE_HTTPS = True

        # app中未设置site host 取设置里的site host
        assertUrlCorrect(settings_host, url_name, req, secure=True)
        
        # app中设置取app设置
        self.app.configurations["SITE_HTTPS"] = False
        self.app.configurations["SITE_HOST"] = configure_host
        assertUrlCorrect(configure_host, url_name, req, secure=False)



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2019-05-25 08:26
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wechat_django', '0004_templatemessage'),
    ]

    operations = [
        migrations.AddField(
            model_name='material',
            name='alias',
            field=models.CharField(blank=True, help_text='\u7d20\u6750\u522b\u540d,\u7528\u4e8e\u7a0b\u5e8f\u5feb\u901f\u67e5\u8be2\u7d20\u6750,\u5355app\u4e0b\u552f\u4e00', max_length=16, null=True, verbose_name='alias'),
        ),
        migrations.AddField(
            model_name='template',
            name='alias',
            field=models.CharField(blank=True, help_text='\u6a21\u677f\u522b\u540d,\u7528\u4e8e\u7a0b\u5e8f\u5feb\u901f\u67e5\u8be2\u6a21\u677f,\u5355app\u4e0b\u552f\u4e00', max_length=16, null=True, verbose_name='alias'),
        ),
        migrations.AddField(
            model_name='wechatuser',
            name='alias',
            field=models.CharField(blank=True, help_text='\u7528\u6237\u522b\u540d,\u7528\u4e8e\u7a0b\u5e8f\u5feb\u901f\u67e5\u8be2\u7528\u6237,\u5355app\u4e0b\u552f\u4e00', max_length=16, null=True, verbose_name='alias'),
        ),
        migrations.AlterField(
            model_name='wechatapp',
            name='type',
            field=models.PositiveSmallIntegerField(choices=[(4, 'MINIPROGRAM'), (0, 'OTHER'), (1, 'SERVICEAPP'), (2, 'SUBSCRIBEAPP')], default=1, verbose_name='type'),
        ),
        migrations.AlterUniqueTogether(
            name='material',
            unique_together=set([('app', 'media_id'), ('app', 'alias')]),
        ),
        migrations.AlterUniqueTogether(
            name='template',
            unique_together=set([('app', 'template_id'), ('app', 'alias')]),
        ),
        migrations.AlterUniqueTogether(
            name='wechatuser',
            unique_together=set([('unionid', 'app'), ('app', 'openid'), ('app', 'alias')]),
        ),
        migrations.AlterField(
            model_name='wechatuser',
            name='nickname',
            field=models.CharField(max_length=32, null=True, verbose_name='nickname'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 2.1.7 on 2019-05-16 07:34
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion

from ..models.permission import upgrade_perms, downgrade_perms


class Migration(migrations.Migration):

    dependencies = [
        ('wechat_django', '0003_usersession'),
    ]

    operations = [
        migrations.CreateModel(
            name='Template',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('template_id', models.CharField(max_length=64, verbose_name='template_id')),
                ('title', models.CharField(max_length=32, verbose_name='title')),
                ('content', models.TextField(verbose_name='content')),
                ('example', models.TextField(blank=True, null=True, verbose_name='example')),
                ('primary_industry', models.CharField(blank=True, max_length=64, null=True, verbose_name='primary industry')),
                ('deputy_industry', models.CharField(blank=True, max_length=64, null=True, verbose_name='deputy industry')),
                ('comment', models.TextField(blank=True, null=True, verbose_name='remark')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('app', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='templates', to='wechat_django.WeChatApp')),
            ],
            options={
                'verbose_name': 'template',
                'verbose_name_plural': 'templates',
            },
        ),
        migrations.AlterUniqueTogether(
            name='template',
            unique_together={('app', 'template_id')},
        ),
    ]

    # 追加两个权限
    migrations.RunPython(
        lambda *args, **kwargs: upgrade_perms("template", "template_sync"),
        lambda *args, **kwargs: downgrade_perms("template", "template_sync"),
    )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import datetime

from django.db import models as m
from django.utils import timezone as tz
import six

from wechat_django.models.base import create_shortcut
from . import WeChatPay


paymethod = create_shortcut(WeChatPay)


class PayDateTimeField(m.DateTimeField):
    _format = "%Y%m%d%H%M%S"
    _tz = tz.pytz.timezone("Asia/Shanghai")

    def to_python(self, value):
        try:
            # 由微信字符串格式转换
            # TODO: 不知道会否有潜在时区问题
            if isinstance(value, six.text_type):
                return self.str2dt(value)
        except ValueError:
            pass
        return super(PayDateTimeField, self).to_python(value)

    def value_to_string(self, obj):
        rv = self.value_from_object(obj)
        return self.dt2str(rv)

    def value_from_object(self, obj):
        rv = super(PayDateTimeField, self).value_from_object(obj)
        return self.fixdt(rv)

    @classmethod
    def fixdt(cls, dt):
        return dt and dt.astimezone(cls._tz) or None

    @classmethod
    def dt2str(cls, dt):
        dt = cls.fixdt(dt)
        return dt.strftime(cls._format) if dt else ""

    @classmethod
    def str2dt(cls, s):
        if not s:
            return None
        dt = tz.datetime.strptime(s, cls._format)
        return tz.make_aware(dt, cls._tz)


class PayBooleanField(m.NullBooleanField):
    def to_python(self, value):
        if value in (True, "Y"):
            return True
        elif value in (False, "N"):
            return False
        return None

    def value_to_string(self, obj):
        val = self.value_from_object(obj)
        if val is True:
            return "Y"
        elif val is False:
            return "N"
        return ""



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import re

from django.db import models as m, transaction
from django.utils import timezone as tz
from django.utils.translation import ugettext_lazy as _
from django.utils.functional import cached_property
from wechatpy.constants import WeChatErrorCode
from wechatpy.exceptions import WeChatClientException

from ..utils.model import enum2choices, model_fields
from ..utils.func import next_chunk
from . import appmethod, WeChatApp, WeChatModel


class WeChatUserManager(m.Manager):
    def upsert_by_dict(self, user_dict, app=None):
        """根据oauth的结果更新"""
        assert "openid" in user_dict, "openid not found"
        app = app or self.instance
        updates = {
            k: v for k, v in user_dict.items()
            if k in model_fields(self.model)
        }
        updates["synced_at"] = tz.now()
        return self.update_or_create(
            defaults=updates, app=app, openid=updates["openid"])[0]


class WeChatUser(WeChatModel):
    class Gender(object):
        UNKNOWN = 0
        MALE = 1
        FEMALE = 2

    class SubscribeScene(object):
        ADD_SCENE_SEARCH = "ADD_SCENE_SEARCH"  # 公众号搜索
        ADD_SCENE_ACCOUNT_MIGRATION = "ADD_SCENE_ACCOUNT_MIGRATION"  # 公众号迁移
        ADD_SCENE_PROFILE_CARD = "ADD_SCENE_PROFILE_CARD"  # 名片分享
        ADD_SCENE_QR_CODE = "ADD_SCENE_QR_CODE"  # 扫描二维码
        ADD_SCENE_PROFILE_LINK = "ADD_SCENEPROFILE LINK"  # 图文页内名称点击
        ADD_SCENE_PROFILE_ITEM = "ADD_SCENE_PROFILE_ITEM"  # 图文页右上角菜单
        ADD_SCENE_PAID = "ADD_SCENE_PAID"  # 支付后关注
        ADD_SCENE_OTHERS = "ADD_SCENE_OTHERS"  # 其他

    app = m.ForeignKey(
        WeChatApp, related_name="users", null=False, editable=False,
        on_delete=m.CASCADE)
    openid = m.CharField(_("openid"), max_length=36, null=False)
    unionid = m.CharField(_("unionid"), max_length=36, null=True)
    alias = m.CharField(
        _("alias"), max_length=16, blank=True, null=True,
        help_text=_("用户别名,用于程序快速查询用户,单app下唯一"))

    nickname = m.CharField(_("nickname"), max_length=32, null=True)
    sex = m.SmallIntegerField(
        _("gender"), choices=enum2choices(Gender), null=True)
    headimgurl = m.CharField(_("avatar"), max_length=256, null=True)
    city = m.CharField(_("city"), max_length=24, null=True)
    province = m.CharField(_("province"), max_length=24, null=True)
    country = m.CharField(_("country"), max_length=24, null=True)
    language = m.CharField(_("language"), max_length=24, null=True)

    subscribe = m.NullBooleanField(_("is subscribed"), null=True)
    subscribe_time = m.IntegerField(_("subscribe time"), null=True)
    subscribe_scene = m.CharField(
        _("subscribe scene"), max_length=32, null=True,
        choices=enum2choices(SubscribeScene))
    qr_scene = m.IntegerField(_("qr scene"), null=True)
    qr_scene_str = m.CharField(_("qr_scene_str"), max_length=64, null=True)

    remark = m.CharField(
        _("WeChat remark"), max_length=30, blank=True, null=True)
    comment = m.TextField(_("remark"), blank=True, null=True)
    groupid = m.IntegerField(_("group id"), null=True)

    created_at = m.DateTimeField(_("created at"), auto_now_add=True)
    updated_at = m.DateTimeField(_("updated at"), auto_now=True)

    synced_at = m.DateTimeField(_("synchronized at"), null=True, default=None)

    objects = WeChatUserManager()

    class Meta(object):
        verbose_name = _("user")
        verbose_name_plural = _("users")

        ordering = ("app", "-created_at")
        unique_together = (
            ("app", "openid"), ("unionid", "app"), ("app", "alias"))

    @property
    def group(self):
        from . import UserTag
        try:
            return self.app.user_tags.get(id=self.groupid).name
        except UserTag.DoesNotExist:
            return None

    def avatar(self, size=132):
        assert size in (0, 46, 64, 96, 132)
        return self.headimgurl and re.sub(r"\d+$", str(size), self.headimgurl)

    @classmethod
    @appmethod
    def user_by_openid(cls, app, openid, ignore_errors=False, sync_user=True):
        """根据用户openid拿到用户对象
        :param ignore_errors: 当库中未找到用户或接口返回失败时还是强行插入user
        :param sync_user: 从服务器拿用户数据
        """
        try:
            return app.users.get(openid=openid)
        except cls.DoesNotExist:
            if sync_user:
                try:
                    return app.fetch_user(openid)
                except Exception as e:
                    if isinstance(e, WeChatClientException)\
                        and e.errcode == WeChatErrorCode.INVALID_OPENID:
                        # 不存在抛异常
                        raise
                    elif ignore_errors:
                        pass # TODO: 好歹记个日志吧...
                    else:
                        raise
            elif not ignore_errors:
                raise
        return app.users.create(openid=openid)

    @classmethod
    @appmethod("sync_users")
    def sync(cls, app, all=False, detail=True):
        """
        :type app: wechat_django.models.WeChatApp
        :param all: 是否重新同步所有用户
        :param detail: 是否同步用户详情
        """
        # 只有重新同步详情的才能全量同步
        all = all and detail

        users = []
        next_openid = not all and app.ext_info.get("last_openid") or None

        iterator = app.client.user.iter_followers(next_openid)
        for openids in next_chunk(iterator):
            users.extend(cls.upsert_users(app, openids, detail))
            # 更新最后更新openid
            app.ext_info["last_openid"] = openids[-1]
            app.save()
        return users

    @classmethod
    @appmethod
    def upsert_users(cls, app, openids, detail=True):
        if detail:
            return cls.fetch_users(app, openids)
        else:
            with transaction.atomic():
                return list(
                    cls.objects.update_or_create(o, **o)[0]
                    for o in map(lambda openid: dict(
                        app=app, openid=openid
                    ), openids)
                )

    @classmethod
    @appmethod
    def fetch_user(cls, app, openid):
        # NotFound重新抛出40003异常
        return cls.fetch_users(app, [openid]).pop()

    @classmethod
    @appmethod
    def fetch_users(cls, app, openids):
        fields = model_fields(cls)
        # TODO: 根据当前语言拉取用户数据
        user_dicts = app.client.user.get_batch(openids)
        update_dicts = map(
            lambda o: {k: v for k, v in o.items() if k in fields},
            user_dicts
        )
        rv = []
        with transaction.atomic():
            tags = list(app.user_tags.all())

            for user_dict in user_dicts:
                defaults = {k: v for k, v in user_dict.items() if k in fields}
                defaults["synced_at"] = tz.now()
                user = cls.objects.update_or_create(
                    defaults=defaults,
                    app=app,
                    openid=defaults["openid"]
                )[0]

                # 拉取标签数据
                tagid_list = user_dict.get("tagid_list")
                if tagid_list:
                    user._tag_local = True
                    user_tags = list(filter(lambda o: o.id in tagid_list, tags))

                    if len(user_tags) != len(tagid_list):
                        # 标签没有完全同步
                        tags = app.sync_usertags()
                        user_tags = list(filter(lambda o: o.id in tagid_list, tags))

                    user.tags.set(user_tags, clear=False)
                    user.save()

                    user._tag_local = False

                rv.append(user)
            return rv

    @classmethod
    def upsert_by_oauth(cls, app, user_dict):
        """根据oauth的结果更新"""
        assert "openid" in user_dict, "openid not found"
        updates = {
            k: v for k, v in user_dict.items()
            if k in model_fields(cls)
        }
        updates["synced_at"] = tz.now()
        return cls.objects.update_or_create(
            defaults=updates, app=app, openid=updates["openid"])[0]

    @cached_property
    def session(self):
        return self.sessions.first()

    def update(self, user_dict=None):
        """重新同步用户数据"""
        if user_dict and isinstance(user_dict, dict):
            # 小程序用户有些字段和公众号不同
            if "avatarUrl" in user_dict:
                user_dict["headimgurl"] = user_dict.pop("avatarUrl")
            if "gender" in user_dict:
                user_dict["sex"] = user_dict.pop("gender")

            field_names = model_fields(self)
            for key in user_dict:
                if key.lower() in field_names:
                    setattr(self, key.lower(), user_dict[key])
            self.save()
        else:
            self.fetch_user(self.app, self.openid)
        self.refresh_from_db()

    def save(self, *args, **kwargs):
        self.alias = self.alias or None
        return super(WeChatUser, self).save(*args, **kwargs)

    def __str__(self):
        return "{nickname}({openid})".format(
            nickname=self.nickname or "", openid=self.openid)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models as m
from django.utils.translation import ugettext_lazy as _
import hashlib
import json
from jsonfield import JSONField

from wechatpy.crypto import WeChatWxaCrypto
from wechatpy.exceptions import InvalidSignatureException

from . import WeChatModel, WeChatUser


class Session(WeChatModel):
    class Type(object):
        MINIPROGRAM = 1

    user = m.ForeignKey(
        WeChatUser, on_delete=m.CASCADE, related_name="sessions", null=False)
    type = m.PositiveSmallIntegerField()
    auth = JSONField(default={})

    created_at = m.DateTimeField(_("created at"), auto_now_add=True)
    updated_at = m.DateTimeField(_("updated at"), auto_now=True)

    @property
    def app(self):
        return self.user.app

    @property
    def app_id(self):
        return self.user.app_id

    @property
    def session_key(self):
        return self.auth.get("session_key")

    def decrypt_message(self, msg, iv):
        """
        :raises: ValueError
        """
        crypto = WeChatWxaCrypto(self.session_key, iv, self.app.appid)
        return crypto.decrypt_message(msg)

    def validate_message(self, msg, sign):
        """
        :raises: wechatpy.exceptions.InvalidSignatureException
        """
        str_to_sign = (msg + self.session_key).encode()
        server_sign = hashlib.sha1(str_to_sign).hexdigest()
        if server_sign != sign:
            raise InvalidSignatureException()
        return json.loads(msg)

    def __str__(self):
        return "session_key: {0}".format(self.session_key)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from hashlib import md5
import json

from django.db import models as m, transaction
from django.utils.translation import ugettext_lazy as _
from jsonfield import JSONField

from ..utils.model import enum2choices
from . import appmethod, MessageHandler, WeChatApp, WeChatModel


class MenuManager(m.Manager):
    def get_queryset(self):
        return (super(MenuManager, self).get_queryset()
            .prefetch_related("sub_button"))


class Menu(WeChatModel):
    class Event(object):
        CLICK = "click"
        VIEW = "view"
        # SCANCODEPUSH = "scancode_push"
        # SCANCODEWAITMSG = "scancode_waitmsg"
        # PICSYSPHOTO = "pic_sysphoto"
        # PICPHOTOORALBUM = "pic_photo_or_album"
        # PICWEIXIN = "pic_weixin"
        # LOCATIONSELECT = "location_select"
        # MEDIAID = "media_id"
        # VIEWLIMITED = "view_limited"

        MINIPROGRAM = "miniprogram"

    app = m.ForeignKey(
        WeChatApp, related_name="menus", on_delete=m.CASCADE)
    name = m.CharField(_("name"), max_length=16)
    menuid = m.IntegerField(_("menuid"), null=True, blank=True)
    parent = m.ForeignKey(
        "Menu", related_name="sub_button", null=True, blank=True,
        on_delete=m.CASCADE)
    type = m.CharField(
        _("type"), max_length=20, choices=enum2choices(Event),
        null=True, blank=True)
    content = JSONField()

    weight = m.IntegerField(_("weight"), default=0, null=False)
    created_at = m.DateTimeField(_("created at"), auto_now_add=True)
    updated_at = m.DateTimeField(_("updated at"), auto_now=True)

    class Meta(object):
        verbose_name = _("menu")
        verbose_name_plural = _("menus")

        ordering = ("app", "-weight", "id")

    objects = MenuManager()

    @classmethod
    @appmethod("sync_menus")
    def sync(cls, app):
        """
        从微信同步菜单数据
        :type app: wechat_django.models.WeChatApp
        """
        resp = app.client.menu.get_menu_info()
        try:
            data = resp["selfmenu_info"]["button"]
        except KeyError:
            data = []

        with transaction.atomic():
            # 旧menu
            app.menus.all().delete()
            # 移除同步菜单产生的message handler
            app.message_handlers.filter(
                src=MessageHandler.Source.MENU).delete()
            rv = [cls.json2menu(menu, app) for menu in data]
        app.ext_info["current_menus"] = cls.menus2json(app)
        app.save()
        return rv

    @classmethod
    @appmethod("publish_menus")
    def publish(cls, app, menuid=None):
        """
        发布菜单
        :type app: wechat_django.models.WeChatApp
        """
        data = cls.menus2json(app, menuid)
        rv = app.client.menu.create(data)
        app.ext_info["current_menus"] = data
        app.save()
        return rv

    @classmethod
    def get_menus(cls, app, menuid=None):
        """获取数据库中公众号菜单配置"""
        q = app.menus.filter(parent_id__isnull=True)
        q = q.filter(menuid=menuid) if menuid else q.filter(menuid__isnull=True)
        return list(q.all())

    @classmethod
    def menus2json(cls, app, menuid=None):
        menus = cls.get_menus(app, menuid)
        return dict(button=[menu.to_json() for menu in menus])

    @classmethod
    def json2menu(cls, data, app):
        """
        :type app: wechat_django.models.WeChatApp
        """
        menu = cls(name=data["name"], app=app)
        menu.type = data.get("type")
        if not menu.type:
            menu.save()
            menu.sub_button.add(*[
                cls.json2menu(sub, app) for sub in
                (data.get("sub_button") or dict(list=[])).get("list")
            ])
        elif menu.type in (
            cls.Event.VIEW, cls.Event.CLICK, cls.Event.MINIPROGRAM):
            menu.content = data
        else:
            # 要当作回复处理了
            menu.type = cls.Event.CLICK
            # 生成一个唯一key
            key = md5(json.dumps(data).encode()).hexdigest()
            menu.content = dict(key=key)
            MessageHandler.from_menu(menu, data)
        menu.save()
        return menu

    def to_json(self):
        rv = dict(name=self.name)
        if self.type:
            rv["type"] = self.type
            if self.type in (Menu.Event.CLICK, Menu.Event.VIEW,
                Menu.Event.MINIPROGRAM):
                rv.update(self.content)
            else:
                raise ValueError("incorrect menu type")
        else:
            rv["sub_button"] = [btn.to_json() for btn in self.sub_button.all()]
        return rv

    def __str__(self):
        return "{0}".format(self.name)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.contrib import admin
from django.utils.translation import ugettext_lazy as _
import object_tool

from ..models import UnifiedOrder, UnifiedOrderResult
from .base import WeChatPayModelAdmin


class OrderResultAdmin(admin.StackedInline):
    model = UnifiedOrderResult

    fields = (
        "transaction_id", "trade_state", "time_end", "settlement_total_fee",
        "cash_fee", "cash_fee_type", "coupon_fee", "bank_type", "detail",
        "is_subscribe", "sub_is_subscribe", "created_at", "updated_at")

    def get_fields(self, request, obj=None):
        rv = super(OrderResultAdmin, self).get_fields(request, obj)
        if not request.app.pay.sub_mch_id:
            rv = list(rv)
            rv.remove("sub_is_subscribe")
        return rv

    def get_readonly_fields(self, request, obj=None):
        return self.get_fields(request, obj)

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class OrderAdmin(WeChatPayModelAdmin):
    __category__ = "pay_order"
    __model__ = UnifiedOrder

    inlines = (OrderResultAdmin,)

    actions = ("sync",)
    list_display = (
        "out_trade_no", "body", "total_fee", "trade_state", "transaction_id",
        "openid", "sub_openid", "time_start", "time_expire", "created_at",
        "updated_at")

    change_object_tools = ("sync",)
    fields = (
        "out_trade_no", "body", "total_fee", "fee_type", "openid",
        "sub_openid", "time_start", "time_expire", "detail",
        "spbill_create_ip", "device_info", "goods_tag", "trade_type",
        "product_id", "limit_pay", "receipt", "scene_info", "comment",
        "created_at", "updated_at")

    def get_list_display(self, request):
        rv = super(OrderAdmin, self).get_list_display(request)
        if not request.app.pay.sub_mch_id:
            rv = list(rv)
            rv.remove("sub_openid")
        return rv

    def get_fields(self, request, obj=None):
        rv = super(OrderAdmin, self).get_fields(request, obj)
        if not request.app.pay.sub_mch_id:
            rv = list(rv)
            rv.remove("sub_openid")
        return rv

    def get_readonly_fields(self, request, obj=None):
        fields = list(self.get_fields(request, obj))
        fields.remove("comment")
        return fields

    def sync(self, request, queryset_or_obj):
        self.check_wechat_permission(request, "change")

        if isinstance(queryset_or_obj, UnifiedOrder):
            queryset_or_obj = [queryset_or_obj]

        for obj in queryset_or_obj:
            obj.sync()
    sync.short_description = _("Sync order")

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False



# -*- coding: utf-8 -*-

"""微信通知处理器"""

from __future__ import unicode_literals

from functools import wraps

from django.conf.urls import include, url
from django.core.exceptions import ObjectDoesNotExist
from django.http import response
from django.utils.translation import ugettext_lazy as _
from django.views.decorators.csrf import csrf_exempt
from wechatpy.exceptions import InvalidSignatureException

from wechat_django.models import WeChatApp
from wechat_django.sites.wechat import BaseWeChatViewSet
from .exceptions import WeChatPayNotifyError


def make_response(msg=None):
    tpl = "<xml><return_msg><![CDATA[{msg}]]></return_msg><return_code><![CDATA[{code}]]></return_code></xml>"
    code = "FAIL" if msg else "SUCCESS"
    xml = tpl.format(code=code, msg=msg)
    return response.HttpResponse(xml, content_type="application/xml")


class NotifyViewSet(BaseWeChatViewSet):
    def notify_view(self, view):
        @csrf_exempt
        @wraps(view)
        def decorated_view(request, appname, payname):
            try:
                pay, data = self._prepare(request, appname, payname)
                return make_response(view(request, pay, data))
            except WeChatPayNotifyError as e:
                return make_response(e.msg)
            except:
                log = WeChatApp.objects.get_by_name(appname).logger("handler")
                log.exception("WeChat Pay notify server error")
                return make_response(_("Internal server error"))

        return decorated_view

    def _prepare(self, request, appname, payname):
        """预处理请求"""
        if request.method != "POST":
            raise WeChatPayNotifyError(_("Method not allowed"))
        xml = request.body
        if not xml:
            raise WeChatPayNotifyError(_("Empty body"))

        try:
            app = self.site.app_queryset.prefetch_related("pays")\
                .get_by_name(appname)
        except WeChatApp.DoesNotExist as e:
            raise WeChatPayNotifyError(_("WeChat application not found"), e)

        if not app.abilities.pay:
            raise WeChatPayNotifyError(_("WeChat pay not configured"))

        try:
            pay = app.pays.get(name=payname)
        except ObjectDoesNotExist as e:
            raise WeChatPayNotifyError(_("WeChat Pay not found"), e)
        try:
            return pay, pay.client.parse_payment_result(xml)
        except InvalidSignatureException as e:
            raise WeChatPayNotifyError(_("Invalid signature"), e)

    def get_urls(self):
        return [
            url(r"^pay/(?P<payname>[-_a-zA-Z\d]+)/notify/", include([
                url(
                    r"^order$",
                    self.notify_view(self.order_notify),
                    name="order_notify"
                )
            ]))
        ]

    def order_notify(self, request, pay, data):
        out_trade_no = data["out_trade_no"]
        try:
            order = pay.orders.get(out_trade_no=out_trade_no)
        except ObjectDoesNotExist:
            return _("Order not found")
        data["trade_state"] = data["result_code"]
        order.update(data)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import re


def in_wechat(request):
    """判断是否时微信环境"""
    ua = request.META["HTTP_USER_AGENT"]
    return bool(re.search(r"micromessenger", ua, re.IGNORECASE))



# -*- coding: utf-8 -*-
# flake8: noqa
from __future__ import unicode_literals

__title__ = "wechat-django"
__description__ = "Django WeChat Extension"
__url__ = "https://github.com/Xavier-Lam/wechat-django"
__version__ = "0.3.0"
__author__ = "Xavier-Lam"
__author_email__ = "Lam.Xavier@hotmail.com"

default_app_config = "wechat_django.apps.WeChatConfig"


from .handler import message_handler
from .oauth import wechat_auth, WeChatOAuthView, WeChatSNSScope



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models as m
from django.utils.translation import ugettext_lazy as _
from jsonfield import JSONField

from wechat_django.models import WeChatModel
from wechat_django.utils.model import enum2choices, model_fields
from ..signals import order_updated
from . import UnifiedOrder
from .base import PayBooleanField, PayDateTimeField


class UnifiedOrderResult(WeChatModel):
    class State(object):
        SUCCESS = "SUCCESS"  # 成功
        REFUND = "REFUND"  # 转入退款
        NOTPAY = "NOTPAY"  # 未支付
        CLOSED = "CLOSED"  # 已关闭
        REVOKED = "REVOKED"  # 已撤销(刷卡支付)
        USERPAYING = "USERPAYING"  # 用户支付中
        PAYERROR = "PAYERROR"  # 支付失败(其他原因，如银行返回失败)
        FAIL = "FAIL"  # 回调拿到失败

    order = m.OneToOneField(
        UnifiedOrder, on_delete=m.CASCADE, related_name="result")

    transaction_id = m.CharField(
        _("transaction_id"), max_length=32, null=True)
    trade_state = m.CharField(
        _("trade_state"), max_length=32, choices=enum2choices(State))
    time_end = PayDateTimeField(_("pay time_end"), null=True)

    settlement_total_fee = m.PositiveIntegerField(
        _("settlement_total_fee"), null=True)
    cash_fee = m.PositiveIntegerField(_("cash_fee"), null=True)
    cash_fee_type = m.CharField(_("cash_fee_type"), max_length=16, null=True)
    coupon_fee = m.PositiveIntegerField(_("coupon_fee"), null=True)

    bank_type = m.CharField(_("bank_type"), max_length=16, null=True)
    detail = m.TextField(_("detail"), max_length=8192, null=True)

    is_subscribe = PayBooleanField(_("is_subscribe"), null=True)
    sub_is_subscribe = PayBooleanField(_("sub_is_subscribe"), null=True)

    ext_info = JSONField(default=dict, editable=False)

    created_at = m.DateTimeField(_("created at"), auto_now_add=True)
    updated_at = m.DateTimeField(_("updated at"), auto_now=True)

    class Meta(object):
        verbose_name = _("Unified order result")
        verbose_name_plural = _("Unified order results")

    def update(self, result, signal=True, verify=True):
        """根据参数更新订单状态"""
        verify and self.order.verify(result)
        # TODO: 支付成功后写入用户
        excludes = ("transaction_id",) if self.transaction_id else tuple()
        all_fields = model_fields(UnifiedOrderResult, excludes=excludes)
        ignore_fields = (
            "return_code", "return_msg", "appid", "mch_id", "device_info",
            "nonce_str", "sign", "result_code", "openid", "trade_type",
            "fee_type", "out_trade_no")
        for k, v in result.items():
            if k not in ignore_fields:
                if k in all_fields:
                    setattr(self, k, v)
                else:
                    self.ext_info[k] = v
        self.save()
        if signal:
            order_updated.send(
                sender=self.order.pay.staticname, result=self, order=self.order,
                state=self.trade_state, attach=result.get("attach"))

    def __str__(self):
        return _("%(order)s reuslt") % dict(order=self.order)



# -*- coding: utf-8 -*-

"""权限模块
约定:
对于manage,article 这种每个公众号拥有的权限,变量名用permission
对于{prefix}{appname}|{perm} 这种permission表中的codename,变量名用perm_name
对于可能携带applabel的codename,变量名亦用perm_name
对于Permission对象,变量名用perm
"""

from __future__ import unicode_literals

from collections import defaultdict
import re

from django.contrib.auth.models import Group, Permission, User
from django.contrib.contenttypes.models import ContentType
from django.db import models as m, transaction
from django.dispatch import receiver
from django.utils.translation import ugettext_lazy as _

from . import WeChatApp

WECHATPERM_PREFIX = "app|"

permissions = {
    "manage": _("Can manage %(appname)s"),
    "article": _("Can edit %(appname)s articles"),
    "article_sync": _("Can sync %(appname)s articles"),
    "material": _("Can edit %(appname)s materials"),
    "material_sync": _("Can sync %(appname)s materials"),
    "menu": _("Can edit %(appname)s menus"),
    "menu_sync": _("Can sync %(appname)s menus"),
    "menu_publish": _("Can publish %(appname)s menus"),
    "messagehandler": _("Can edit %(appname)s message handlers"),
    "messagehandler_sync": _("Can sync %(appname)s message handlers"),
    "messagelog": _("Can view %(appname)s message logs"),
    "user": _("Can edit %(appname)s users"),
    "user_sync": _("Can sync %(appname)s users"),
    "usertag": _("Can edit %(appname)s user tags"),
    "usertag_sync": _("Can sync %(appname)s user tags"),
    "template": _("Can edit %(appname)s templates"),
    "template_sync": _("Can sync %(appname)s templates"),
}

permission_required = {
    "material": (
        "article",
        "wechat_django.delete_material"
    ),
    "article": (
        "wechat_django.delete_article",
    ),
    "menu": (
        "wechat_django.delete_menu",
    ),
    "messagehandler": (
        "wechat_django.delete_messagehandler",
        "wechat_django.add_reply",
        "wechat_django.change_reply",
        "wechat_django.delete_reply",
        "wechat_django.add_rule",
        "wechat_django.change_rule",
        "wechat_django.delete_rule"
    )
}


def list_perm_names(app):
    """列举该app下的所有权限"""
    rv = set()
    rv.add(get_perm_name(app))
    rv.update(
        get_perm_name(app, permission)
        for permission in permissions
    )
    return rv


def get_perm_name(app, permission=""):
    if not permission:
        return "{prefix}{appname}".format(
            prefix=WECHATPERM_PREFIX,
            appname=app.name
        )
    return "{prefix}{appname}|{permission}".format(
        prefix=WECHATPERM_PREFIX,
        appname=app.name,
        permission=permission
    )


def get_perm_names(app, permissions):
    return tuple(get_perm_name(app, permission) for permission in permissions)


def get_perm_desc(perm_name, app):
    appname, permission = match_permission(perm_name)
    desc = permissions[permission] if permission\
        else _("Can full control %(appname)s")
    return "{0} | {1}".format(appname, desc) % dict(appname=app.name)


def get_require_perm_names(appname, permission=None):
    """获取依赖的django权限"""
    rv = set()
    perms = (permission_required.get(permission, []) if permission
        else permission_required.keys())
    for perm in perms:
        if perm.find(".") >= 0:
            rv.add(perm)
        else:
            rv.update(get_require_perm_names(appname, perm))
    return rv


def get_perm_model(perm_name):
    """由permission获取permission model"""
    try:
        applabel, codename = perm_name.split(".")
    except:
        applabel = "wechat_django"
        codename = perm_name
    return Permission.objects.get(
        codename=codename,
        content_type__app_label=applabel
    )


def get_perms_by_codenames(codenames):
    return (Permission.objects
        .filter(content_type__app_label__startswith="wechat_django")
        .filter(codename__in=codenames)
        .all())


@receiver(m.signals.post_save, sender=WeChatApp)
def create_app_perms(sender, instance, created, *args, **kwargs):
    if created:
        # 添加
        content_type = ContentType.objects.get_for_model(WeChatApp)
        Permission.objects.bulk_create(
            Permission(
                codename=perm_name,
                name=get_perm_desc(perm_name, instance),
                content_type=content_type
            )
            for perm_name in list_perm_names(instance)
        )


@receiver(m.signals.post_delete, sender=WeChatApp)
def delete_app_perms(sender, instance, *args, **kwargs):
    content_type = ContentType.objects.get_for_model(WeChatApp)
    Permission.objects.filter(
        content_type=content_type,
        codename__in=list_perm_names(instance)
    ).delete()


@receiver(m.signals.m2m_changed, sender=Group.permissions.through)
@receiver(m.signals.m2m_changed, sender=User.user_permissions.through)
def add_required_perms(sender, instance, action, *args, **kwargs):
    if action == "pre_add":
        if isinstance(instance, User):
            perms_set = instance.user_permissions
        else:
            perms_set = instance.permissions
        # 检查相关权限
        perm_names = set()
        perms = Permission.objects.filter(id__in=kwargs["pk_set"]).all()
        for perm in perms:
            appname, permission = match_permission(perm.codename)
            if appname:
                perm_names.update(
                    get_require_perm_names(appname, permission))
        if perm_names:
            codenames = list(map(lambda o: o.split(".")[1], perm_names))
            perms_set.add(*get_perms_by_codenames(codenames))


def get_user_permissions(user, app=None, exclude_sub=False, exclude_manage=False):
    """列举用户所有的微信权限
    :type user: django.contrib.auth.models.User
    """
    excludes = set()
    if exclude_manage:
        excludes.add("manage")
    if exclude_sub:
        excludes.update(
            permission
            for permission in permissions
            if permission.find("_") != -1
        )

    perm_names = user.get_all_permissions()
    rv = defaultdict(set)
    for perm_name in perm_names:
        appname, permission = match_permission(perm_name)
        if appname:
            if permission:
                rv[appname].add(permission)
            else:
                # 所有权限
                rv[appname] = set(permissions.keys())

    for appname in rv:
        rv[appname].difference_update(excludes)

    return rv[app.name] if app else dict(rv.items())


def match_permission(perm_name):
    """从permission的codename中拿到appname与权限名"""
    str4format = r"(?:{label}[.])?{prefix}(?P<appname>.+?)(?:|(?P<permission>.+))?$"
    pattern = str4format.format(
        label="wechat_django",
        prefix=WECHATPERM_PREFIX
    ).replace("|", "[|]")
    match = re.match(pattern, perm_name)
    if match:
        return match.group("appname"), match.group("permission")
    else:
        return None, None


def upgrade_perms(permissions):
    """迁移时新增权限"""
    with transaction.atomic():
        content_type = ContentType.objects.get_for_model(WeChatApp)
        apps = WeChatApp.objects.all()
        for app in apps:
            for perm_name in get_perm_names(app, permissions):
                query = dict(
                    codename=perm_name,
                    content_type=content_type
                )
                defaults = query.copy()
                defaults.update(dict(name=get_perm_desc(perm_name, app)))
                Permission.objects.update_or_create(
                    defaults=defaults, **query)


def downgrade_perms(permissions):
    """降级时移除新增的权限"""
    with transaction.atomic():
        content_type = ContentType.objects.get_for_model(WeChatApp)
        apps = WeChatApp.objects.all()
        for app in apps:
            Permission.objects.filter(
                content_type=content_type,
                codename__in=get_perm_names(app, permissions)
            ).delete()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
import object_tool

from ...models import Article
from ..utils import anchor
from ..base import WeChatModelAdmin


class ArticleAdmin(WeChatModelAdmin):
    __category__ = "article"
    __model__ = Article

    changelist_object_tools = ("sync",)
    list_display = ("title", "author", "material_link", "index", "digest",
        "link", "source_url", "synced_at")
    list_editable = ("index",)
    search_fields = ("title", "author", "digest", "content_source_url", "content")

    fields = ("title", "author", "digest", "thumb_image", "thumb_media_id",
        "link", "show_cover_pic", "_content", "content_source_url", "source_url")
    readonly_fields = fields

    link = anchor(_("link"), lambda self, obj: obj.url)
    link.short_description = _("link")

    source_url = anchor(_("source_url"),
        lambda self, obj: obj.content_source_url)
    source_url.short_description = _("source_url")

    @mark_safe
    def thumb_image(self, obj):
        return obj.thumb_url and '<a href="{0}"><img width="200" src="{0}" /></a>'.format(obj.thumb_url)
    thumb_image.short_description = _("thumb_url")

    @mark_safe
    def material_link(self, obj):
        m = obj.material
        if m:
            return '<a href="{link}" title="{title}">{title}</a>'.format(
                link=reverse(
                    "admin:wechat_django_material_change",
                    kwargs=dict(
                        object_id=m.id,
                        wechat_app_id=m.app_id
                    )
                ),
                title="{comment} ({media_id})".format(
                    comment=m.comment,
                    media_id=m.media_id
                ) if m.comment else m.media_id
            )
    material_link.short_description = _("material")

    @mark_safe
    def _content(self, obj):
        return obj.content
    _content.short_description = _("content")

    def get_queryset(self, request):
        base_q = super(WeChatModelAdmin, self).get_queryset(request)
        return base_q.filter(material__app_id=request.app_id)

    def get_actions(self, request):
        actions = super(ArticleAdmin, self).get_actions(request)
        if 'delete_selected' in actions:
            del actions['delete_selected']
        return actions

    @object_tool.confirm(short_description=_("Sync articles"))
    def sync(self, request, obj=None):
        self.check_wechat_permission(request, "sync")

        def action():
            materials = request.app.sync_articles()
            msg = _("%(count)d articles successfully synchronized")
            return msg % dict(count=len(materials))
        
        return self._clientaction(
            request, action, _("Sync articles failed with %(exc)s"))

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def get_model_perms(self, request):
        if not request.app.abilities.material:
            return {}
        return super(ArticleAdmin, self).get_model_perms(request)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import threading
from uuid import uuid4

from ..utils.func import next_chunk, Static
from .base import WeChatTestCase


class UtilFunctoolTestCase(WeChatTestCase):
    def test_next_chunk(self):
        """测试next chunk"""
        data = list(next_chunk(range(50)))
        self.assertEqual(len(data), 1)
        self.assertEqual(set(range(50)), set(data[0]))

        data = list(next_chunk(range(100)))
        self.assertEqual(len(data), 1)
        self.assertEqual(set(range(100)), set(data[0]))

        data = list(next_chunk(range(101), 100))
        self.assertEqual(len(data), 2)
        self.assertEqual(set(range(100)), set(data[0]))
        self.assertEqual(set((100, )), set(data[1]))

    def test_static(self):
        """测试static"""
        total = 10
        strings = [str(uuid4()) for i in range(total)]
        another_strings = [s.encode().decode() for s in strings]

        for i in range(total):
            # 验证直接id不相等
            self.assertEqual(strings[i], another_strings[i])
            self.assertNotEqual(id(strings[i]), id(another_strings[i]))
            self.assertEqual(
                id(Static(strings[i])), id(Static(another_strings[i])))

        # 在另一个线程也正常
        def another_thread(strings):
            another_strings = [s.encode().decode() for s in strings]

            for i in range(total):
                # 验证直接id不相等
                self.assertEqual(strings[i], another_strings[i])
                self.assertNotEqual(id(strings[i]), id(another_strings[i]))
                self.assertEqual(
                    id(Static(strings[i])), id(Static(another_strings[i])))

        threading.Thread(target=another_thread, args=(strings,))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import datetime

from django.db import models as m
from django.dispatch import receiver
from django.utils import timezone as tz
from django.utils.translation import ugettext_lazy as _
from jsonfield import JSONField
from wechatpy.pay.utils import get_external_ip

from wechat_django.models import WeChatModel
from wechat_django.utils.model import enum2choices, model_fields
from wechat_django.utils.web import get_ip
from . import WeChatPay
from .base import PayDateTimeField, paymethod


class UnifiedOrder(WeChatModel):
    ALLOW_UPDATES = (
        "device_info", "openid", "sub_openid", "trade_type", "total_fee",
        "fee_type")

    class TradeType(object):
        JSAPI = "JSAPI"
        NATIVE = "NATIVE"
        APP = "APP"
        MWEB = "MWEB"
        MICROPAY = "MICROPAY"

    class LimitPay(object):
        NONE = None
        NOCREDIT = "no_credit"

    class Receipt(object):
        NONE = None
        TRUE = "Y"

    class FeeType(object):
        CNY = "CNY"

    pay = m.ForeignKey(WeChatPay, on_delete=m.CASCADE, related_name="orders")

    device_info = m.CharField(_("device_info"), max_length=32, null=True)
    body = m.CharField(_("order body"), max_length=128)
    detail = m.TextField(_("order detail"), max_length=6000, null=True)
    out_trade_no = m.CharField(_("out_trade_no"), max_length=32)
    fee_type = m.CharField(
        _("fee_type"), max_length=16, null=True, default=FeeType.CNY)
    total_fee = m.PositiveIntegerField(_("total_fee"))
    spbill_create_ip = m.CharField(_("spbill_create_ip"), max_length=64) # TODO: 生成时是否应该赋值?
    time_start = PayDateTimeField(_("time_start"), default=tz.now)
    time_expire = PayDateTimeField(_("time_expire"))
    goods_tag = m.CharField(_("goods_tag"), max_length=32, null=True)
    trade_type = m.CharField(
        _("trade_type"), max_length=16, choices=enum2choices(TradeType))
    product_id = m.CharField(_("product id"), max_length=32, null=True)
    limit_pay = m.CharField(
        _("limit_pay"), max_length=32, null=True, default=LimitPay.NONE,
        choices=enum2choices(LimitPay))
    openid = m.CharField(_("openid"), max_length=128, null=True)
    sub_openid = m.CharField(_("sub openid"), max_length=128, null=True)
    receipt = m.CharField(
        _("recept"), max_length=8, null=True, default=Receipt.NONE,
        choices=enum2choices(Receipt))
    scene_info = JSONField(_("scene_info"), max_length=256, null=True)

    comment = m.TextField(_("comment"), blank=True)

    ext_info = JSONField(default=dict, editable=False)
    _call_args = JSONField(
        db_column="call_args", default=dict, null=True, editable=False)

    created_at = m.DateTimeField(_("created at"), auto_now_add=True)
    updated_at = m.DateTimeField(_("updated at"), auto_now=True)

    class Meta(object):
        verbose_name = _("Unified order")
        verbose_name_plural = _("Unified orders")

        index_together = (("pay", "created_at"), )
        unique_together = (("pay", "out_trade_no"),)

    @property
    def app_id(self):
        return self.pay.app_id

    def transaction_id(self):
        try:
            return self.result.transaction_id
        except AttributeError:
            return None
    transaction_id.short_description = _("transaction_id")

    def trade_state(self):
        try:
            return self.result.trade_state
        except AttributeError:
            return None
    trade_state.short_description = _("trade_state")

    @classmethod
    @paymethod("create_order")
    def create(cls, pay, user=None, request=None, **kwargs):
        """
        :type pay: wechat_django.pay.models.WeChatPay
        :param user: 发生支付的用户,如不填写,请填写openid或sub_openid
        :type user: wechat_django.models.WeChatUser
        :param request: 发生此次请求的request
        :param kwargs: 参见模型其他各字段的值
        """
        data = {}

        if user:
            if pay.sub_appid and user.app.appid == pay.sub_appid:
                data["sub_openid"] = user.openid
            else:
                data["openid"] = user.openid

        # 自动构造trade_type
        data["trade_type"] = cls.TradeType.JSAPI

        fields = model_fields(UnifiedOrder)
        data.update({k: v for k, v in kwargs.items() if k in fields})

        return cls.objects.create(pay=pay, **data)

    def call_args(self, request=None, dt2py=False, **kwargs):
        """调用统一下单接口的参数
        :param attach: 附加数据
        :param kwargs: 覆盖默认生成的数据
        :param dt2py: 将请求微信的datetime格式转换为python的datetime格式
        """
        # 更新ip
        self.spbill_create_ip = kwargs.pop(
            "client_ip",
            get_ip(request) or self.spbill_create_ip or get_external_ip())
        if self._call_args:
            self._call_args["client_ip"] = self.spbill_create_ip
        else:
            self.time_start = kwargs.pop("time_start", self.time_start)
            self.time_expire = kwargs.pop("time_expire", self.time_expire)
            # 显式设置起止时间
            if not self.time_start:
                self.time_start = self.created_at
            if not self.time_expire:
                self.time_expire = self.created_at + datetime.timedelta(hours=2)
            # 构建call_args并缓存
            notify_url = self.pay.app.build_url(
                "order_notify", request=request,
                kwargs=dict(payname=self.pay.name), absolute=True)
            self._call_args = dict(
                trade_type=self.trade_type,
                body=self.body,
                total_fee=self.total_fee,
                notify_url=notify_url,
                client_ip=self.spbill_create_ip,
                user_id=self.openid,
                out_trade_no=self.out_trade_no,
                detail=self.detail,
                fee_type=self.fee_type,
                time_start=PayDateTimeField.dt2str(self.time_start),
                time_expire=PayDateTimeField.dt2str(self.time_expire),
                goods_tag=self.goods_tag,
                product_id=self.product_id,
                device_info=self.device_info,
                limit_pay=self.limit_pay,
                scene_info=self.scene_info,
                sub_user_id=self.sub_openid,
                receipt=self.receipt
            )
            self._call_args.update(kwargs)
        self.save()

        rv = self._call_args.copy()
        if dt2py:
            # 将时间重新to_python
            rv["time_start"] = PayDateTimeField.str2dt(rv["time_start"])
            rv["time_expire"] = PayDateTimeField.str2dt(rv["time_expire"])
        return rv

    def prepay(self, request=None, **kwargs):
        """调用统一下单接口"""
        call_args = self.call_args(request, dt2py=True, **kwargs)
        return self.pay.client.order.create(**call_args)

    def jsapi_params(self, prepay_id, *args, **kwargs):
        return self.pay.client.jsapi.get_jsapi_params(
            prepay_id, *args, **kwargs)

    def close(self):
        """关闭订单"""
        rv = self.pay.client.order.close(out_trade_no=self.out_trade_no)
        self.sync()
        return rv

    def reverse(self):
        """撤销订单"""
        rv = self.pay.client.order.reverse(out_trade_no=self.out_trade_no)
        self.sync()
        return rv

    def sync(self):
        """更新订单状态"""
        result = self.pay.client.order.query(
            out_trade_no=self.out_trade_no)
        self.update(result)
        return self.result, result

    def update(self, result, signal=True, verify=True):
        """由字典更新数据"""
        verify and self.verify(result)
        updated = False
        for field in self.ALLOW_UPDATES:
            data = result.get(field)
            if not getattr(self, field) and data:
                setattr(self, field, data)
                updated = True
        updated and self.save()

        try:
            self.result.update(result, signal=signal, verify=False)
        except AttributeError:
            from . import UnifiedOrderResult
            transaction_id = result.get("transaction_id") or None
            obj = UnifiedOrderResult(
                order=self, transaction_id=transaction_id)
            obj.update(result, signal=signal, verify=False)

    def verify(self, result):
        """检查订单结果参数"""
        from . import UnifiedOrderResult
        # 检查mch_id是否一致
        assert result["mch_id"] == self.pay.mch_id, "incorrect mch_id"

        # 检查out_order_no是否一致
        assert result["out_trade_no"] == self.out_trade_no,\
            "incorrect out trade no"

        # 检查金额一致
        if result["trade_state"] == UnifiedOrderResult.State.SUCCESS:
            assert str(result["total_fee"]) == str(self.total_fee),\
                "incorrect total fee"

    def __str__(self):
        return "{0} ({1})".format(self.out_trade_no, self.body)


@receiver(m.signals.pre_save, sender=UnifiedOrder)
def on_order_created(sender, instance, *args, **kwargs):
    """订单保存前初始化订单参数"""
    if instance.time_expire is None:
        # 订单过期时间默认创建时间2小时后
        instance.time_expire = instance.time_start + tz.timedelta(hours=2)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django import forms
from django.contrib import admin
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
import object_tool

from ...models import UserTag, WeChatUser
from ..base import WeChatModelAdmin


class UserForm(forms.ModelForm):
    tags = forms.ModelMultipleChoiceField(
        UserTag.objects.all(),
        widget=admin.widgets.FilteredSelectMultiple(_("tags"), False),
        required=False
    )

    def __init__(self, *args, **kwargs):
        super(UserForm, self).__init__(*args, **kwargs)
        if self.instance.pk:
            self.initial["tags"] = self.instance.tags.values_list(
                "pk", flat=True)

    def save(self, *args, **kwargs):
        instance = super(UserForm, self).save(*args, **kwargs)
        if instance.pk:
            instance.tags.set(self.cleaned_data["tags"], clear=False)
        return instance


class WeChatUserAdmin(WeChatModelAdmin):
    __category__ = "user"
    __model__ = WeChatUser

    actions = ("update",)
    changelist_object_tools = ("incremental_sync", "sync_all")
    list_display = (
        "openid", "nickname", "avatar", "alias", "subscribe", "remark",
        "created_at")
    search_fields = ("=openid", "=unionid", "alias", "nickname", "remark")

    fields = (
        "avatar", "nickname", "openid", "unionid", "alias", "sex", "city",
        "province", "country", "language", "subscribe", "subscribetime",
        "subscribe_scene", "qr_scene", "qr_scene_str", "remark", "comment",
        "tags", "group", "created_at", "updated_at")

    @mark_safe
    def avatar(self, obj):
        return obj.headimgurl and '<img src="{0}" />'.format(obj.avatar(46))
    avatar.short_description = _("avatar")
    avatar.allow_tags = True

    def subscribetime(self, obj):
        return obj.subscribe_time and timezone.datetime.fromtimestamp(obj.subscribe_time)
    subscribetime.short_description = _("subscribe time")

    def sync(self, request, obj=None, method="sync", kwargs=None):
        self.check_wechat_permission(request, "sync")
        kwargs = kwargs or dict()
        # 可能抛出48001 没有api权限
        def action():
            users = getattr(WeChatUser, method)(request.app, **kwargs)
            msg = _("%(count)d users successfully synchronized")
            return msg % dict(count=len(users))

        tpl = _("%(method)s failed with %(exc)s")
        return self._clientaction(
            request, action, tpl, dict(method=_(method)))

    @object_tool.confirm(short_description=_("Incremental sync users"))
    def incremental_sync(self, request, obj=None):
        return self.sync(request, obj)

    @object_tool.confirm(short_description=_("Sync all users"))
    def sync_all(self, request, obj=None):
        return self.sync(request, obj, kwargs=dict(all=True))

    update = lambda self, request, queryset: self.sync(
        request, None, "fetch_users", dict(
            openids=[o.openid for o in queryset.all()]
        ))
    update.short_description = _("update selected")

    def get_actions(self, request):
        actions = super(WeChatUserAdmin, self).get_actions(request)
        if 'delete_selected' in actions:
            del actions['delete_selected']
        if not request.app.abilities.user_manager:
            del actions["update"]
        return actions

    def get_object_tools(self, request):
        rv = super(WeChatUserAdmin, self).get_object_tools(request)
        if not request.app.abilities.user_manager:
            rv.pop("incremental_sync", None)
            rv.pop("sync_all", None)
        return rv

    form = UserForm

    def get_form(self, request, obj=None, **kwargs):
        # 过滤只显示自己app的标签 并过滤系统标签
        app = request.app
        form = super(WeChatUserAdmin, self).get_form(request, obj, **kwargs)
        tags_field = form.declared_fields["tags"]

        tags_field.queryset = (tags_field.queryset
            .filter(app=app)
            .exclude(id__in=UserTag.SYS_TAGS))
        return form

    def get_readonly_fields(self, request, obj=None):
        return tuple(o for o in self.fields if o not in (
            "alias", "remark", "tags", "comment"))

    def save_model(self, request, obj, form, change):
        if "remark" in form.changed_data:
            obj.app.client.user.update_remark(obj.openid, obj.remark)
        return super(WeChatUserAdmin, self).save_model(
            request, obj, form, change)

    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2019-05-01 08:52
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('wechat_django', '0002_usertags'),
    ]

    operations = [
        migrations.CreateModel(
            name='Session',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type', models.PositiveSmallIntegerField()),
                ('auth', jsonfield.fields.JSONField(default={})),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='updated at')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='sessions', to='wechat_django.WeChatUser')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AlterModelOptions(
            name='usertag',
            options={'ordering': ('app', 'id'), 'verbose_name': 'user tag', 'verbose_name_plural': 'user tags'},
        ),
        migrations.AlterField(
            model_name='wechatapp',
            name='type',
            field=models.PositiveSmallIntegerField(choices=[(4, 'MINIPROGRAM'), (1, 'SERVICEAPP'), (2, 'SUBSCRIBEAPP')], default=1, verbose_name='type'),
        ),
        migrations.AlterUniqueTogether(
            name='usertag',
            unique_together=set([('app', 'id')]),
        ),
        migrations.AlterIndexTogether(
            name='usertag',
            index_together=set([('app', 'name')]),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import re
from setuptools import find_packages, setup
import sys


package = dict()

with open(os.path.join("wechat_django", "__init__.py"), "r") as f:
    lines = f.readlines()
    for line in lines:
        match = re.match(r"(__\w+?__)\s*=\s*(.+)$", line)
        if match:
            package[match.group(1)] = eval(match.group(2))

with open("README.md", "rb") as f:
    long_description = f.read().decode("utf-8")

with open("requirements.txt") as f:
    requirements = [l for l in f.read().splitlines() if l]

keywords = [
    "WeChat", "weixin", "wx", "WeChatPay", "micromessenger", "django",
    "微信", "微信支付"]
if sys.version_info.major == 2:
    keywords.remove("微信")
    keywords.remove("微信支付")

setup(
    name=package["__title__"],
    version=package["__version__"],
    author=package["__author__"],
    author_email=package["__author_email__"],
    url=package["__url__"],
    packages=find_packages(),
    keywords=",".join(keywords),
    description=package["__description__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    include_package_data=True,
    tests_require=requirements + ["cryptography>=2.5", "httmock==1.2.6"],
    test_suite="runtests.main",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Web Environment",
        "Framework :: Django :: 1.11",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: Chinese (Simplified)",
        "Natural Language :: Chinese (Traditional)",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: Site Management",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities"
    ],
    extras_require={
        "cryptography": ["cryptography"],
        "pycrypto": ["pycryptodome"],
    }
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from wechatpy.client.api import WeChatMessage, WeChatTemplate, WeChatWxa

from ..models import Template, WeChatUser
from .base import mock, WeChatTestCase


class TemplateTestCase(WeChatTestCase):
    def test_sync_miniprogram(self):
        """测试同步小程序模板"""
        with mock.patch.object(WeChatWxa, "list_templates"):
            templates = [{
                "template_id": "wDYzYZVxobJivW9oMpSCpuvACOfJXQIoKUm0PY397Tc",
                "title": "购买成功通知",
                "content": "购买地点{{keyword1.DATA}}\n购买时间{{keyword2.DATA}}\n物品名称{{keyword3.DATA}}\n",
                "example": "购买地点：TIT造舰厂\n购买时间：2016年6月6日\n物品名称：咖啡\n"
            }]
            WeChatWxa.list_templates.return_value = templates

            # 同步
            self.miniprogram.sync_templates()
            self.assertTemplateEqual(self.miniprogram, templates)

            # 再次同步 结果一致
            self.miniprogram.sync_templates()
            self.assertTemplateEqual(self.miniprogram, templates)

    def test_sync_service(self):
        """测试同步服务号模板"""
        with mock.patch.object(WeChatTemplate, "get_all_private_template"):
            data = self.load_data("get_all_private_template")
            templates = data["template_list"]
            WeChatTemplate.get_all_private_template.return_value = data

            # 同步
            self.app.sync_templates()
            self.assertTemplateEqual(self.app, templates)

            # 再次同步 结果一致
            self.app.sync_templates()
            self.assertTemplateEqual(self.app, templates)

    def test_send_miniprogram(self):
        """测试发送小程序模板消息"""
        with mock.patch.object(WeChatWxa, "send_template_message"):
            id = "id"
            openid = "openid"
            form_id = "form_id"
            pagepath = "pagepath"
            emphasis_keyword = "keyword1.DATA"
            data = {
                "keyword1": {
                    "value": "339208499"
                },
                "keyword2": {
                    "value": "2015年01月05日 12:30"
                },
                "keyword3": {
                    "value": "粤海喜来登酒店"
                } ,
                "keyword4": {
                    "value": "广州市天河区天河路208号"
                }
            }

            t = Template(app=self.miniprogram, template_id=id)

            # 直接发送openid及data
            t.send(openid, data, form_id=form_id)
            self.assertCallArgsEqual(
                WeChatWxa.send_template_message, (openid, id, data, form_id))

            # 发送user及data
            user = WeChatUser(openid=openid)
            t.send(user, data, form_id=form_id)
            self.assertCallArgsEqual(
                WeChatWxa.send_template_message, (openid, id, data, form_id))

            # 发送data及page
            t.send(
                openid, data, form_id=form_id, pagepath=pagepath,
                emphasis_keyword=emphasis_keyword)
            self.assertCallArgsEqual(
                WeChatWxa.send_template_message, (openid, id, data, form_id),
                {"page": pagepath, "emphasis_keyword": emphasis_keyword})

            # 发送kwargs
            t.send(
                openid, form_id=form_id,
                **{k: v["value"] for k, v in data.items()})
            self.assertCallArgsEqual(
                WeChatWxa.send_template_message, (openid, id, data, form_id))

    def test_send_service(self):
        """测试发送服务号模板消息"""
        with mock.patch.object(WeChatMessage, "send_template"):
            id = "id"
            openid = "openid"
            url = "url"
            data = {
                "first": {
                    "value":"恭喜你购买成功！",
                    "color":"#173177"
                },
                "keyword1":{
                    "value":"巧克力",
                    "color":"#173177"
                },
                "keyword2": {
                    "value":"39.8元",
                    "color":"#173177"
                },
                "keyword3": {
                    "value":"2014年9月22日",
                    "color":"#173177"
                },
                "remark":{
                    "value":"欢迎再次购买！",
                    "color":"#173177"
                }
           }

            t = Template(app=self.app, template_id=id)

            # 发送data及url
            t.send(openid, data, url)
            self.assertCallArgsEqual(
                WeChatMessage.send_template, (openid, id, data), {"url": url})

            # 使用user对象
            user = WeChatUser(openid=openid)
            t.send(user, data)
            self.assertCallArgsEqual(
                WeChatMessage.send_template, (openid, id, data))

            # 发送data及小程序
            appid = "appid"
            pagepath = "pagepath"
            miniprogram = dict(
                appid=appid,
                pagepath=pagepath
            )
            t.send(openid, data, **miniprogram)
            self.assertCallArgsEqual(
                WeChatMessage.send_template, (openid, id, data),
                {"mini_program": miniprogram})

            # 发送kwargs
            t.send(openid, **data)
            self.assertCallArgsEqual(
                WeChatMessage.send_template, (openid, id, data))

            # 发送str kwargs
            t.send(openid, **{k: v["value"] for k, v in data.items()})
            self.assertCallArgsEqual(
                WeChatMessage.send_template,
                (openid, id, {k: {"value": v["value"]} for k, v in data.items()}))

    def assertTemplateEqual(self, app, templates):
        """模板和同步的一致"""
        self.assertEqual(app.templates.count(), len(templates))
        for item in templates:
            template = app.templates.get(
                template_id=item["template_id"])
            for key, value in item.items():
                self.assertEqual(getattr(template, key), value)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class WeChatPayConfig(AppConfig):
    name = "wechat_django.pay"
    label = "wechat_django_pay"
    verbose_name = _("WeChat Pay")
    verbose_name_plural = _("WeChat Pay")

    def ready(self):
        # 处理微信支付相关权限
        import wechat_django.models.permission as base_pm
        from .models.permisssion import permissions, permission_required
        base_pm.permissions.update(permissions)
        base_pm.permission_required.update(permission_required)

        # 注册viewsets
        from wechat_django.sites import wechat
        from .notify import NotifyViewSet
        wechat.register(NotifyViewSet)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.utils.translation import ugettext_lazy as _
import object_tool

from ...models import Template
from ..base import WeChatModelAdmin


class TemplateAdmin(WeChatModelAdmin):
    __category__ = "template"
    __model__ = Template

    actions = None
    changelist_object_tools = ("sync",)
    list_display = ("template_id", "alias", "title", "content", "created_at")
    search_fields = (
        "template_id", "alias", "title", "content", "primary_industry",
        "deputy_industry", "comment")

    fields = (
        "template_id", "alias", "title", "content", "primary_industry",
        "deputy_industry", "example", "comment", "created_at")
    readonly_fields = (
        "template_id", "title", "content", "primary_industry",
        "deputy_industry", "example", "created_at")

    @object_tool.confirm(short_description=_("Sync templates"))
    def sync(self, request, obj=None):
        self.check_wechat_permission(request, "sync")

        def action():
            templates = request.app.sync_templates()
            msg = _("%(count)d templates successfully synchronized")
            return msg % dict(count=len(templates))

        return self._clientaction(
            request, action, _("Sync templates failed with %(exc)s"))
    sync.short_description = _("sync templates")

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def get_model_perms(self, request):
        if not request.app.abilities.template:
            return {}
        return super(TemplateAdmin, self).get_model_perms(request)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models as m, transaction
from django.utils.translation import ugettext_lazy as _

from . import appmethod, Material, WeChatModel


class Article(WeChatModel):
    material = m.ForeignKey(
        Material, related_name="articles", on_delete=m.CASCADE)

    title = m.CharField(_("title"), max_length=64)
    thumb_media_id = m.CharField(_("thumb_media_id"), max_length=64)
    show_cover_pic = m.BooleanField(_("show cover"), default=True)
    author = m.CharField(
        _("author"), max_length=24, blank=True, null=True, default="")
    digest = m.CharField(
        _("digest"), max_length=256, blank=True, null=True, default="")
    content = m.TextField(_("content"))
    url = m.CharField(_("url"), max_length=256)
    content_source_url = m.CharField(
        _("content source url"), max_length=256)

    need_open_comment = m.NullBooleanField(
        _("need open comment"), default=None)
    only_fans_can_comment = m.NullBooleanField(
        _("only fans can comment"), default=None)

    index = m.PositiveSmallIntegerField(_("index"))
    _thumb_url = m.CharField(
        db_column="thumb_url", max_length=256, null=True, default=None)

    synced_at = m.DateTimeField(_("synchronized at"), auto_now_add=True)

    class Meta(object):
        verbose_name = _("article")
        verbose_name_plural = _("articles")

        unique_together = (("material", "index"),)
        ordering = ("material", "index")

    @property
    def app(self):
        return self.material and self.material.app

    @property
    def app_id(self):
        return self.material and self.material.app_id

    @property
    def thumb_url(self):
        if self._thumb_url is None and self.thumb_media_id:
            # 不存在url时通过thumb_media_id同步
            image = Material.objects.filter(
                app=self.app, media_id=self.thumb_media_id).first()
            self._thumb_url = image and image.url
            self._thumb_url is not None and self.save()
        return self._thumb_url

    @thumb_url.setter
    def thumb_url(self, value):
        self._thumb_url = value

    @classmethod
    @appmethod("sync_articles")
    def sync(cls, app, id=None):
        if id:
            return Material.sync(app, id, Material.Type.NEWS)
        else:
            with transaction.atomic():
                return Material.sync_type(app, Material.Type.NEWS)

    def to_json(self):
        return dict(
            title=self.title,
            thumb_media_id=self.thumb_media_id,
            show_cover_pic=1 if self.show_cover_pic else 0,
            need_open_comment=1 if self.need_open_comment else 0,
            only_fans_can_comment=1 if self.only_fans_can_comment else 0,
            author=self.author,
            digest=self.digest,
            content=self.content,
            url=self.url,
            content_source_url=self.content_source_url
        )

    def __str__(self):
        return "{0}".format(self.title)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json

from wechatpy.exceptions import InvalidSignatureException

from ..models import Session, WeChatUser
from .base import mock, WeChatTestCase


class UserTestCase(WeChatTestCase):
    def test_sync(self):
        """测试同步用户"""
        pass

    def test_fetch_users(self):
        """测试拉取用户"""
        pass

    def test_upsert_users(self):
        """测试插入或更新用户"""
        pass

    def test_update(self):
        """测试更新用户"""
        pass

    def test_miniprogram_messages(self):
        """测试小程序消息解析"""
        self.app.appid = "wx4f4bc4dec97d474b"
        session_key = "HyVFkGl5F5OQWJZZaNzBBg=="
        user = WeChatUser.objects.create(openid="openid", app=self.app)
        session = Session(
            user=user,
            type=Session.Type.MINIPROGRAM,
            auth=dict(session_key=session_key)
        )

        # 验证签名
        raw_data = '{"nickName":"Band","gender":1,"language":"zh_CN","city":"Guangzhou","province":"Guangdong","country":"CN","avatarUrl":"http://wx.qlogo.cn/mmopen/vi_32/1vZvI39NWFQ9XM4LtQpFrQJ1xlgZxx3w7bQxKARol6503Iuswjjn6nIGBiaycAjAtpujxyzYsrztuuICqIM5ibXQ/0"}' # noqa
        sign = "75e81ceda165f4ffa64f4068af58c64b8f54b88c"
        data = session.validate_message(raw_data, sign)
        self.assertEqual(data, json.loads(raw_data))
        sign = "fake_sign"
        self.assertRaises(
            InvalidSignatureException, session.validate_message,
            raw_data, sign)

        # 解密
        session.auth["session_key"] = "tiihtNczf5v6AKRyjwEUhQ=="
        encrypted_data = "CiyLU1Aw2KjvrjMdj8YKliAjtP4gsMZMQmRzooG2xrDcvSnxIMXFufNstNGTyaGS9uT5geRa0W4oTOb1WT7fJlAC+oNPdbB+3hVbJSRgv+4lGOETKUQz6OYStslQ142dNCuabNPGBzlooOmB231qMM85d2/fV6ChevvXvQP8Hkue1poOFtnEtpyxVLW1zAo6/1Xx1COxFvrc2d7UL/lmHInNlxuacJXwu0fjpXfz/YqYzBIBzD6WUfTIF9GRHpOn/Hz7saL8xz+W//FRAUid1OksQaQx4CMs8LOddcQhULW4ucetDf96JcR3g0gfRK4PC7E/r7Z6xNrXd2UIeorGj5Ef7b1pJAYB6Y5anaHqZ9J6nKEBvB4DnNLIVWSgARns/8wR2SiRS7MNACwTyrGvt9ts8p12PKFdlqYTopNHR1Vf7XjfhQlVsAJdNiKdYmYVoKlaRv85IfVunYzO0IKXsyl7JCUjCpoG20f0a04COwfneQAGGwd5oa+T8yO5hzuyDb/XcxxmK01EpqOyuxINew==" # noqa
        iv = "r7BXXKkLb8qrSNn05n0qiA=="
        decrypted = session.decrypt_message(encrypted_data, iv)
        self.assertEqual(self.app.appid, decrypted["watermark"]["appid"])
        fake_encrypted_data = "abc"
        self.assertRaises(
            (TypeError, ValueError), session.decrypt_message,
            fake_encrypted_data, iv)
        fake_iv = "abc"
        self.assertRaises(
            (TypeError, ValueError), session.decrypt_message, encrypted_data,
            fake_iv)
        exipred_session_key = "abc"
        session.auth["session_key"] = exipred_session_key
        self.assertRaises(
            (TypeError, ValueError), session.decrypt_message, encrypted_data,
            iv)

    def test_update_by_user_dict(self):
        """小程序用户数据更新"""
        openid = "openid"
        user = WeChatUser.objects.create(app=self.app, openid=openid)
        origin_dict = {
            "nickName": "Band",
            "gender": 1,
            "language": "zh_CN",
            "city": "Guangzhou",
            "province": "Guangdong",
            "country": "CN",
            "avatarUrl": "http://wx.qlogo.cn/mmopen/vi_32/1vZvI39NWFQ9XM4LtQpFrQJ1xlgZxx3w7bQxKARol6503Iuswjjn6nIGBiaycAjAtpujxyzYsrztuuICqIM5ibXQ/0"
        }
        user_dict = origin_dict.copy()
        user.update(user_dict)
        self.assertEqual(user.nickname, origin_dict["nickName"])
        self.assertEqual(user.sex, origin_dict["gender"])
        self.assertEqual(user.language, origin_dict["language"])
        self.assertEqual(user.city, origin_dict["city"])
        self.assertEqual(user.province, origin_dict["province"])
        self.assertEqual(user.country, origin_dict["country"])
        self.assertEqual(user.headimgurl, origin_dict["avatarUrl"])
